#%% -*- coding: utf-8 -*-
"""
Created on Sat Jun 20 15:53:32 2020

@author: Katharina Rath
"""
import numpy as np
from scipy.optimize import fmin, minimize, newton
from func import (bluenoise, integrate_pendulum, nlp, nlpreg, buildK, build_K, plot_pendulum,
                  buildKreg, applymap, nlp_with_grad, nlp, nlpreg, nlpreg_with_grad, nll, nll_chol)
from param import  sig, Nm, N, dtsymp, sig_n, qmin, qmax, pmin, pmax, qminmap, qmaxmap, pminmap, pmaxmap, Ntest
import matplotlib.pyplot as plt
from profit.util.halton import halton
# from profit.sur.backend.gp_functions import nll, nll_chol
from init import xtrain, ytrain, ztrain, q, p, P, t, ztrain1,ztrain2
from sklearn.metrics import mean_squared_error
import cma
from GPy import *

#%%
#Definition of the parameters:

# Lengthscale:
l = [4.801629659374284, 4.801629659374284]

# sigma_noise:
sig_n_opt = [2.8294119107878543e-4]

#%%
#Build K(x,x):

# Custom:
K = np.empty((2*len(xtrain), 2*len(xtrain)))
buildK(xtrain, ytrain, xtrain, ytrain, l, K)
plt.figure(1)
plt.imshow(K)

# GPy:
sq_kern = kern.RBF(2,variance=sig**2,lengthscale=l[0]) # the kernel function

x1= np.array([xtrain,ytrain]).T

K1 = sq_kern.dK2_dXdX2(x1, x1, 0, 0) # upper left part of K
K2 = sq_kern.dK2_dXdX2(x1, x1, 0, 1) # upper right part of K
K3 = sq_kern.dK2_dXdX2(x1, x1, 1, 0) # lower left part of K
K4 = sq_kern.dK2_dXdX2(x1, x1, 1, 1) # lower right part of K

K5 = np.block([[K1, K2],[K3, K4]]) # the whole matrix K (or L)
plt.figure(2)
plt.imshow(K5)

# Comparision:
print('Comparaision of K and K5: ',np.max(np.abs(K-K5)), '\n')

#%%
#Build the inverse:

# Custom:
Kyinv = np.linalg.pinv(K + sig_n_opt[0]**2*np.eye(K.shape[0]))
plt.figure(3)
plt.imshow(Kyinv)

# GPy:
Kyinv5 = np.linalg.pinv(K5 + sig_n_opt[0]**2*np.eye(K5.shape[0]))
plt.figure(4)
plt.imshow(Kyinv5)

# Comparision:
print('Comparaison of Kyinv and Kyinv5: ',np.max(np.abs(Kyinv-Kyinv5)), '\n')

#%%
sq_kern1 = kern.RBF(1,variance=sig**2,lengthscale=l[0])
sq_kern2 = kern.DiffKern(sq_kern1, 0)

gauss_likelihood = likelihoods.Gaussian(variance=2.8294119107878543e-4**2)
# Build the model (GPy):
# m = models.GPRegression( x1 , np.array([ztrain1,ztrain2]).T , sq_kern, noise_var = sig_n_opt[0]**2 )
#m = models.MultioutputGP(X_list=[xtrain.reshape((-1,1)), ytrain.reshape((-1,1))], Y_list=[ztrain1.reshape((-1,1)), ztrain2.reshape((-1,1))], kernel_list=[sq_kern2, sq_kern2], likelihood_list = [gauss_likelihood,gauss_likelihood])

# models.MultioutputGP([xtrain.reshape((-1,1)), ytrain.reshape((-1,1))], [ztrain1.reshape((-1,1)), ztrain2.reshape((-1,1))],
#     kernel_list = [K1, K4], kernel_cross_covariances = {(0,1): K2, (1,0): K3},
#     likelihood_list = [gauss_likelihood,gauss_likelihood])
# Display the model:
print(m)

#%%
#check how well interpolation is done

#test data:

# The number of test points:
nxtest = 4
nytest = 4
Nt = nxtest*nytest

# Definition of the test points:
xt1 = np.linspace(qmin,qmax,nxtest)
yt1 = np.linspace(pmin,pmax,nytest)
xtest, ytest = np.meshgrid(xt1, yt1)
xtest = xtest.T
ytest = ytest.T

ysinttest = integrate_pendulum(xtest.flatten(), ytest.flatten(), t)

plt.figure()
for ik in range(Nt):
    plt.plot(ysinttest[ik][0], ysinttest[ik][1], '.')

Ptest = np.empty((Nt))
Qtest = np.empty((Nt))

for ik in range(0, Nt):
    Ptest[ik] = ysinttest[ik][1,-1]
    Qtest[ik] = ysinttest[ik][0,-1]

#%%
#Build Kstar:

# Custom:
Kstar = np.empty((2*Nt, 2*len(xtrain)))
buildK(xtrain, ytrain, xtest.flatten(), Ptest.flatten(), l, Kstar)
plt.figure()
plt.imshow(Kstar)

# GPy:
x2 = np.array([xtest.flatten(),Ptest]).T

Kstar1 = sq_kern.dK2_dXdX2(x1, x2, 0, 0) # upper left part of Kstar
Kstar2 = sq_kern.dK2_dXdX2(x1, x2, 0, 1) # upper right part of Kstar
Kstar3 = sq_kern.dK2_dXdX2(x1, x2, 1, 0) # lower left part of Kstar
Kstar4 = sq_kern.dK2_dXdX2(x1, x2, 1, 1) # lower right part of Kstar

Kstar5 = np.block([[Kstar1, Kstar2],[Kstar3, Kstar4]]) # the whole matrix Kstar (or Lstar)
Kstar5 = Kstar5.T
plt.figure()
plt.imshow(Kstar5)

# Comparaision:
print('Comparaison of Kstar and Kstar5: ',np.max(np.abs(Kstar-Kstar5)), '\n')

#%%
#Mean of the prediction:

# Custom:
Ef = Kstar.dot(Kyinv.dot(ztrain))

# GPy
# mean and variance matrix star :
# Ef5, varf5 = m.predict(np.array([xtest.flatten(),Ptest.flatten()]).T, full_cov=False)
Ef5, varf5 = m.predict([xtest.flatten().reshape((-1,1)),Ptest.flatten().reshape((-1,1))], full_cov=False)
Ef5 = Ef5.flatten()

# Comparaison
print('Comparaison of Ef and Ef5: ',np.max(np.abs(Ef-Ef5)), '\n')

#%%
#Definition of (deltap , deltaq):

# Custom:
deltap = Ef[0:Nt] - (ytest.flatten() - Ptest.flatten())
deltaq = Ef[Nt:2*Nt] - (Qtest.flatten() - xtest.flatten())

dist = np.sqrt(np.sum(np.hstack((deltap, deltaq))**2, axis = 0)/(2*Nt))
print('dist: ',dist, '\n')
plt.figure()
plt.plot(Ef[0:Nt], Ef[Nt:2*Nt], '.')
plt.plot(ytest.flatten() - Ptest.flatten(), (Qtest.flatten() - xtest.flatten()), 'x')

z1test = ytest.flatten() - Ptest.flatten()
z2test = Qtest.flatten() - xtest.flatten()
ztest =  np.concatenate((z1test, z2test))
outtest = mean_squared_error(ztest, Ef)
print('outtest: ', outtest, '\n')

# GPy:
deltap5 = Ef5[0:Nt] - (ytest.flatten() - Ptest.flatten())
deltaq5 = Ef5[Nt:2*Nt] - (Qtest.flatten() - xtest.flatten())

dist5 = np.sqrt(np.sum(np.hstack((deltap5, deltaq5))**2, axis = 0)/(2*Nt))
print('dist5: ',dist5, '\n')
plt.figure()
plt.plot(Ef5[0:Nt], Ef5[Nt:2*Nt], '.')
plt.plot(ytest.flatten() - Ptest.flatten(), (Qtest.flatten() - xtest.flatten()), 'x')

outtest5 = mean_squared_error(ztest, Ef5)
print('outest5: ', outtest5, '\n')

#%%
# symplectic mapping

#set initial conditions (q0, p0) on grid (taken from config file)
q0map = np.linspace(qminmap,qmaxmap,Ntest)
p0map = np.linspace(pminmap,pmaxmap,Ntest)
[Q0map,P0map] = np.meshgrid(q0map,p0map)
Q0map = Q0map.transpose()
P0map = P0map.transpose()

# fit second GP + hyperparameter optimization to have a first guess for newton for P
xtrainp = q.flatten()
ytrainp = p.flatten()
ztrainp = P.flatten()
lp = [0.2, 0.2]

sol_kreg = minimize(nlpreg_with_grad, l,
    method='L-BFGS-B', args = (xtrainp, ytrainp, ztrainp, sig_n), jac=True,
    tol=1e-6, options={'xtol':1e-6})

lp = sol_kreg.x

#%%
# build K and its inverse:

# Custom:
Kyp = buildKreg(xtrainp, ytrainp, lp)
Kyinvp = np.linalg.pinv(Kyp + sig_n**2*np.eye(Kyp.shape[0]))

# GPy:
sq_kern1 = kern.RBF(2,variance=sig**2,lengthscale=lp[0]) # the kernel function
Kyp5 = sq_kern1.K(np.array([xtrainp, ytrainp]).T,np.array([xtrainp, ytrainp]).T)
Kyinvp5 = np.linalg.pinv(Kyp5 + sig_n**2*np.eye(Kyp5.shape[0]))

# Comparaison:
print('Comparaison of Kyp and Kyp5: ', np.max(np.abs(Kyp-Kyp5)), '\n')
print('Comparaison of Kyinvp and Kyinvp5: ', np.max(np.abs(Kyinvp-Kyinvp5)), '\n')

#%% apply map
# pmaptest is added for debugging
# and should be equal to pmap  -> however, it isn't

# Custom:
qmap, pmap, H, pmaptest = applymap(
    l, lp, Q0map, P0map, xtrainp, ytrainp, ztrainp, Kyinvp, xtrain, ytrain, ztrain, Kyinv)

# GPy:
qmap5, pmap5, H5, pmaptest5 = applymap(
    l, lp, Q0map, P0map, xtrainp, ytrainp, ztrainp, Kyinvp5, xtrain, ytrain, ztrain, Kyinv5)


#%%
# Custom:
np.savez('data',qmap= qmap, pmap=pmap, H=H)

print('qmap.shape: ', qmap.shape, '\n')

import matplotlib.pyplot as plt

data = np.load('data.npz')
qmap = data['qmap']
pmap = data['pmap']
H = data['H']

qplot = qmap[0]
pplot = pmap[0]
q0 = qmap[0][:,0]
p0 = pmap[0][0]

# GPy:
np.savez('data5',qmap= qmap5, pmap=pmap5, H=H5)

print('qmap5.shape: ', qmap5.shape, '\n')

import matplotlib.pyplot as plt

data5 = np.load('data5.npz')
qmap5 = data['qmap']
pmap5 = data['pmap']
H5 = data['H']

qplot5 = qmap[0]
pplot5 = pmap[0]
q05 = qmap[0][:,0]
p05 = pmap[0][0]

#%%

# Custom:
plt.figure()
for i in range(0, qmap.shape[2]):
    plt.plot(qmap[:,:,i], pmap[:,:,i], 'k,')
plt.plot(q, p, 'kx')
plt.plot(qplot, pplot, 'ko')
plt.xlabel('q')
plt.ylabel('p')
# plt.title('mapped data')
plt.savefig('dataGP.png')
plt.figure()
for i in range(0, qmap.shape[2]):
    for j in range(0, qmap.shape[1]):
        plt.plot(H[:,j,i]/(np.mean(H[:,j,i])))
plt.xlabel('t')
plt.ylabel('H')
plt.title('energyGP')
plt.savefig('energyGP.png')

# GPy:
plt.figure()
for i in range(0, qmap5.shape[2]):
    plt.plot(qmap5[:,:,i], pmap5[:,:,i], 'k,')
plt.plot(q, p, 'kx')
plt.plot(qplot5, pplot5, 'ko')
plt.xlabel('q5')
plt.ylabel('p5')
# plt.title('mapped data')
plt.savefig('dataGP5.png')
plt.figure()
for i in range(0, qmap5.shape[2]):
    for j in range(0, qmap5.shape[1]):
        plt.plot(H5[:,j,i]/(np.mean(H5[:,j,i])))
plt.xlabel('t5')
plt.ylabel('H5')
plt.title('energyGP5')
plt.savefig('energyGP5.png')


"""
Created on Wed Aug 28 01:47:54 2020
@author: manal khallaayoune
"""

import numpy as np
import GPy
from GPy.kern.src.stationary import Stationary
from cosine_prod import Cosine_prod
import func_ProdExtended
import matplotlib.pyplot as plt

def test_kern_ProdExtended():

    # Training points for 1D:
    x0train = np.linspace(-5,5,100).reshape(-1,1)
    x1train = np.linspace(-2,2,100).reshape(-1,1)
    x2train = np.linspace(0,9,100).reshape(-1,1)
    x3train = np.linspace(8,18,100).reshape(-1,1)
    
    #Training points for 2D:
    xtrain = np.hstack((x0train, x1train))
    
    ##Training points for 3D:
    #xtrain = np.hstack((x0train, x1train, x2train))
    
    ##Training points for 4D:
    #xtrain = np.hstack((x0train, x1train, x2train, x3train))

    # 2D kernel:
    k1 = Cosine_prod(1, active_dims = 0)
    k2 = Cosine_prod(1, active_dims = 1)
    k0 = ProdExtended((k1,k2))
    k0_K = k0.K(xtrain,xtrain)
    dk0 = k0.dK_dX(xtrain, xtrain, 0)
    # dk0 = k0.dK_dX(xtrain, xtrain, 1)
    dk0_2 = k0.dK2_dXdX2(xtrain, xtrain, 0, 0)
    # dk0_2 = k0.dK2_dXdX2(xtrain, xtrain, 0, 1)
    # dk0_2 = k0.dK2_dXdX2(xtrain, xtrain, 1, 0)
    # dk0_2 = k0.dK2_dXdX2(xtrain, xtrain, 1, 1)
    
    # 2D test:
    k3 = Cosine_prod(2)
    K = k3.K(xtrain,xtrain)
    dK = k3.dK_dX(xtrain, xtrain, 0)
    # dK = k3.dK_dX(xtrain, xtrain, 1)
    dK2 = k3.dK2_dXdX2(xtrain, xtrain, 0, 0)
    # dK2 = k3.dK2_dXdX2(xtrain, xtrain, 0, 1)
    # dK2 = k3.dK2_dXdX2(xtrain, xtrain, 1, 0)
    # dK2 = k3.dK2_dXdX2(xtrain, xtrain, 1, 1)
    
    # 2D plot :
    # plot the matrices
    plt.figure()
    plt.imshow(K)
    plt.title('K sympy' '\n' '2D')
    plt.figure()
    plt.imshow(k0_K)
    plt.title('K GPy' '\n' '2D')
    # plot one of the courbes
    plt.figure()
    plt.plot(K[49,:])
    plt.title('K sympy' '\n' '2D')
    plt.figure()
    plt.plot(k0_K[49,:])
    plt.title('K GPy' '\n' '2D')
    
    plt.figure()
    plt.imshow(dK)
    plt.title('1st derivative sympy' '\n' '2D')
    plt.figure()
    plt.imshow(dk0)
    plt.title('1st derivative GPy' '\n' '2D')
    plt.figure()
    plt.plot(dK[49,:])
    plt.title('1st derivative sympy' '\n' '2D')
    plt.figure()
    plt.plot(dk0[49,:])
    plt.title('1st derivative GPy' '\n' '2D')
    
    plt.figure()
    plt.imshow(dK2)
    plt.title('2nd derivative sympy' '\n' '2D')
    plt.figure()
    plt.imshow(dk0_2)
    plt.title('2nd derivative GPy' '\n' '2D')
    plt.figure()
    plt.plot(dK2[49,:])
    plt.title('2nd derivative sympy' '\n' '2D')
    plt.figure()
    plt.plot(dk0_2[49,:])
    plt.title('2nd derivative GPy' '\n' '2D')

    # Comparision (any dimension) :
    print(np.isclose(k0_K, K, rtol=1e-6),'\n')
    print(np.isclose(dk0, dK, rtol=1e-6),'\n')
    print(np.isclose(dk0_2, dK2, rtol=1e-6))
    
    

test_kern_ProdExtended()



"""
Created: Mon Jul  8 11:47:38 2019
@author: Christopher Albert <albert@alumni.tugraz.at>
"""

import numpy as np
import matplotlib.pyplot as plt
import os

from collections import OrderedDict
from chaospy import J, generate_quadrature, orth_ttr, fit_quadrature, E, Std, descriptives, Normal

from profit.sur.gaussian_process import GPySurrogate

import pickle

#%%

def convert():
  indata = np.genfromtxt('run_1998-2003_new/1997-2001_V7_input_full.txt', skip_header=1)
  np.save('indata.npy', indata)
  
  outpath = 'run_1998-2003_new/FULL/'
  fnames = os.listdir(outpath)
  outdata = np.genfromtxt(os.path.join(outpath, fnames[0]), skip_header=1)
  for fname in fnames[1:]:
    data = np.genfromtxt(os.path.join(outpath, fname), skip_header=1)
    outdata = np.concatenate([outdata, data])
  np.save('outdata.npy', outdata)
  
#%%
def fit():
  nodes, weights = generate_quadrature(4, distribution, rule='G', sparse=False)
  print(np.max(nodes - indata.T))
  expansion = orth_ttr(3, distribution)
  return fit_quadrature(expansion, nodes, weights, outdata)

#%% save to disk
def save(approx):
  with open('pce.pickle', 'wb') as filehandler:
      pickle.dump(approx, filehandler)

#%%
indata = np.load('indata.npy')
outdata = np.load('outdata.npy')
with open('pce.pickle', 'rb') as filehandler:
  approx = pickle.load(filehandler)

params = OrderedDict()
data = np.genfromtxt('params_1998.txt', dtype=None, encoding=None)
for param in data:
    if(param[1] == 'Normal'):
        params[param[0]] = Normal(param[2], param[3])

#%%
distribution = J(*params.values())

#%%
F0 = E(approx, distribution)
dF = Std(approx, distribution)

#%%
F0[0] = F0[1]
dF[0] = dF[1]
plt.plot(np.arange(len(F0))-38, F0, 'k')
plt.fill_between(np.arange(len(F0))-38,F0-dF, F0+dF, color='k', alpha=0.3)
plt.ylim(-100,200)
plt.xlabel('t')
plt.ylabel('f(t)')
plt.show()

#%%
sobol1 = descriptives.sensitivity.Sens_m(approx, distribution)

#%%
plt.figure()
plt.plot(np.arange(len(F0))-38, sobol1.T)
plt.legend(params)
plt.title('First order sensitivity indices (Sobol)')
plt.xlabel('t')
plt.ylabel('f(t)')

#%% Surrogate over x

nx = 38
ns = 30
ntrain = nx*ns

samples = distribution.sample(ns)
y = ((approx(*samples)[41:].T)/dF[41:]).T


plt.figure()
plt.plot(y[5,:], 'x')
plt.show()

xtrain = 1.0*np.repeat(np.arange(nx), ns).reshape([ntrain, 1])
ytrain = y.reshape([ntrain, 1])
#%%
sur = GPySurrogate()
sur.train(xtrain, ytrain)

#%%
ntest = 200
xtest = np.linspace(0, 44, ntest).reshape([ntest, 1])
ftest = sur.predict(xtest)

plt.figure()
plt.plot(xtrain, ytrain, '.', markersize=1)
plt.plot(np.arange(38), F0[41:]/dF[41:], 'k')
plt.plot(xtest, ftest[0], color='tab:blue')
plt.fill_between(xtest[:,0], (ftest[0][:,0] - 1.96*np.sqrt(ftest[1][:,0])), 
                             (ftest[0][:,0] + 1.96*np.sqrt(ftest[1][:,0])),
                 alpha = 0.3)
plt.plot(np.arange(38), F0[41:]/dF[41:]-1.96, 'k--')
plt.plot(np.arange(38), F0[41:]/dF[41:]+1.96, 'k--')
plt.xlabel('t')
plt.ylabel('f(t)')
#plt.ylim(-100,200)
plt.legend(['samples', 'reference', 'surrogate'])
plt.show()
#%%

plt.plot(F0, 'k')
plt.fill_between(range(len(F0)),F0-dF, F0+dF)
plt.ylim(-100,200)
plt.show()



r"""This module contains the backends for various Gaussian Process surrogate models.

Gaussian Processes (GPs) are a generalization of Gaussian Distributions which are described by mean- and covariance
functions.
They can be used as as a non-parametric, supervised machine-learning technique for regression and classification.
The advantages of GPs over other machine-learning techniques such as Artificial Neural Networks is the consistent,
analytic mathematical derivation within probability-theory and therefore their intrinsic uncertainty-quantification of
the predictions by means of the covariance function. This makes the results of GP fits intuitive to interpret.

GPs belong (like e.g. Support Vector Machines) to the kernel methods of machine learning. The mean function is often
neglected and set to zero, because most functions can be modelled with a suitable covariance function or a combination
of several covariance functions. The most important kernels are the Gaussian (RBF) and its generalization,
the Matern kernels.

Isotropic RBF Kernel:
$$
\begin{align}
k(x, x') &= \sigma_f^2 \exp(-\frac{1}{2} \frac{\lvert x-x' \rvert}{l^2})
\end{align}
$$

Literature:
    Rasmussen & Williams 2006: General Introduction to Gaussian Processes
    Garnett 2014: Active Learning and Linear Embeddings
    Osborne 2012: Active Learning of Hyperparameters
"""

from abc import ABC, abstractmethod
import numpy as np
from .sur import Surrogate


class GaussianProcess(Surrogate, ABC):
    r"""This is the base class for all Gaussian Process models.

    Attributes:
        trained (bool): Flag that indicates if the model is already trained and ready to make predictions.
        fixed_sigma_n (bool/float/ndarray): Indicates if the data noise should be optimized or not.
            If an ndarray is given, its length must match the training data.
        Xtrain (ndarray): Input training points.
        ytrain (ndarray): Observed output data.
            Vector output is supported for independent variables only.
        ndim (int): Dimension of input data.
        output_ndim (int): Dimension of output data.
        kernel (str/object): Kernel identifier such as 'RBF' or directly the (surrogate specific) kernel object.
            Defaults to 'RBF'.
        hyperparameters (dict): Parameters like length-scale, variance and noise which can be optimized during training.
            As default, they are inferred from the training data.

    Default parameters:
        surrogate: GPy
        kernel: RBF

    Default hyperparameters:
        $l$ ... length scale
        $\sigma_f$ ... scaling
        $\sigma_n$ ... data noise

        $$
        \begin{align}
        l &= \frac{1}{2} \overline{\lvert x - x' \rvert} \\
        \sigma_f &= \overline{std(y)} \\
        \sigma_n &= 0.01 \cdot \overline{max(y) - min(y)}
        \end{align}
        $$
    """

    _defaults = {'surrogate': Surrogate._defaults['surrogate'],  # Default parameters for all GP surrogates
                 'kernel': 'RBF',
                 'hyperparameters': {'length_scale': None,  # Hyperparameters are inferred from training data
                                     'sigma_n': None,
                                     'sigma_f': None}}

    def __init__(self):
        super().__init__()
        self.kernel = None
        self.hyperparameters = {}

    def prepare_train(self, X, y, kernel=None, hyperparameters=None, fixed_sigma_n=False, multi_output=False):
        """Check the training data, initialize the hyperparameters and set the kernel either from the given parameter,
        from config or from the default values.

        Parameters:
            X: (n, d) array of input training data.
            y: (n, D) array of training output.
            kernel (str/object): Identifier of kernel like 'RBF' or directly the kernel object of the
                specific surrogate.
            hyperparameters (dict): Hyperparameters such as length scale, variance and noise.
                Taken either from given parameter, config file or inferred from the training data.
                The hyperparameters can be different depending on the kernel. E.g. The length scale can be a scalar,
                a vector of the size of the training data, or for the custom LinearEmbedding kernel a matrix.
            fixed_sigma_n (bool/float/ndarray): Indicates if the data noise should be optimized or not.
                If an ndarray is given, its length must match the training data.
            multi_output (bool): Indicates if a multi output is desired if y has more than one dimension.
                Otherwise the excess dimensions are used as independent variable.
        """

        self.Xtrain = np.atleast_2d(X)
        self.ytrain = np.atleast_2d(y)
        self.encode_training_data()
        self.ndim = self.Xtrain.shape[-1]
        self.multi_output = self.multi_output or multi_output
        self.output_ndim = self.ytrain.shape[-1] if self.multi_output else 1
        self.fixed_sigma_n = self.fixed_sigma_n or fixed_sigma_n

        # Infer hyperparameters from training data
        inferred_hyperparameters = {'length_scale': np.array([0.5*np.mean(
            np.linalg.norm(self.Xtrain[:, None, :] - self.Xtrain[None, :, :], axis=-1))]),
                                    'sigma_f': np.array([np.mean(np.std(self.ytrain, axis=0))]),
                                    'sigma_n': np.array([1e-2*np.mean(self.ytrain.max(axis=0) - self.ytrain.min(axis=0))])}

        # Set hyperparameters either from config, the given parameter or inferred from the training data
        self.hyperparameters = self.hyperparameters or hyperparameters or self._defaults['hyperparameters'].copy()
        for key, value in self.hyperparameters.items():
            if value is None:
                value = inferred_hyperparameters[key]
            self.hyperparameters[key] = np.atleast_1d(value)
        self.print_hyperparameters('Initial')

        # Set kernel from config, parameter or default and convert a string to the class object
        self.kernel = self.kernel or kernel or self._defaults['kernel']
        if isinstance(self.kernel, str):
            self.kernel = self.select_kernel(self.kernel)

    @abstractmethod
    def train(self, X, y, kernel=None, hyperparameters=None, fixed_sigma_n=False, return_hess_inv=False,
              multi_output=False):
        """Trains the model on the dataset.

        After initializing the model with a kernel function and initial hyperparameters,
        it can be trained on input data X and observed output data y by optimizing the model's hyperparameters.
        This is done by minimizing the negative log likelihood.

        Parameters:
            X (ndarray): (n, d) array of input training data.
            y (ndarray): (n, D) array of training output.
            kernel (str/object): Identifier of kernel like 'RBF' or directly the kernel object of the surrogate.
            hyperparameters (dict): Hyperparameters such as length scale, variance and noise.
                Taken either from given parameter, config file or inferred from the training data.
                The hyperparameters can be different depending on the kernel. E.g. The length scale can be a scalar,
                a vector of the size of the training data, or for the custom LinearEmbedding kernel a matrix.
            fixed_sigma_n (bool): Indicates if the data noise should be optimized or not.
            return_hess_inv (bool): Whether to the attribute hess_inv after optimization. This is important
                for active learning.
            multi_output (bool): Indicates if a multi output is desired if y has more than one dimension.
                Otherwise the excess dimensions are used as independent variable.
        """
        pass

    def prepare_predict(self, Xpred):
        """Prepares the surrogate for prediction by checking if it is trained and validating the data.

        Parameters:
            Xpred (ndarray): Input points for prediction

        Returns:
            ndarray: Checked input data or default values inferred from training data.
        """

        if not self.trained:
            raise RuntimeError("Need to train() before predict()!")

        if Xpred is None:
            Xpred = self.default_Xpred()
        Xpred = np.atleast_2d(Xpred)
        Xpred = self.encode_predict_data(Xpred)
        return Xpred

    @abstractmethod
    def predict(self, Xpred, add_data_variance=True):
        r"""Predicts the output at test points Xpred.

        Parameters:
            Xpred (ndarray/list): Input points for prediction.
            add_data_variance (bool): Adds the data noise $\sigma_n^2$ to the prediction variance.
                This is especially useful for plotting.

        Returns:
            tuple: a tuple containing:
                - ymean (ndarray) Predicted output values at the test input points.
                - yvar (ndarray): Diagonal of the predicted covariance matrix.
        """
        pass

    @abstractmethod
    def optimize(self, **opt_kwargs):
        """Find optimized hyperparameters of the model. Optional kwargs for tweaking optimization.

        Parameters:
            opt_kwargs: Keyword arguments for optimization.
        """
        pass

    @classmethod
    def from_config(cls, config, base_config):
        """Instantiate a GP model from the configuration file with kernel and hyperparameters.

        Parameters:
            config (dict): Only the 'fit' part of the base_config.
            base_config (dict): The whole configuration parameters.

        Returns:
            profit.sur.gaussian_process.GaussianProcess: Instantiated surrogate.
        """

        self = cls()
        self.kernel = config['kernel']
        self.hyperparameters = config['hyperparameters']
        return self

    @classmethod
    def handle_subconfig(cls, config, base_config):
        """Sets default parameters if not existent.

        Does this recursively for each element of the hyperparameter dict.

        Parameters:
            config (dict): Only the 'fit' part of the base_config.
            base_config (dict): The whole configuration parameters.
        """

        for key, default in cls._defaults.items():
            if key not in config:
                config[key] = default
            else:
                if isinstance(default, dict):
                    for kkey, ddefault in default.items():
                        if kkey not in config[key]:
                            config[key][kkey] = ddefault

    @abstractmethod
    def select_kernel(self, kernel):
        """Convert the name of the kernel as string to the kernel class object of the surrogate.

        Parameters:
            kernel (str): Kernel string such as 'RBF' or depending on the surrogate also product and sum kernels
                such as 'RBF+Matern52'.
        Returns:
            object: Custom or imported kernel object. This is the function which builds the kernel and not
            the calculated covariance matrix.
        """
        pass

    def print_hyperparameters(self, prefix):
        """Helper function to print the hyperparameter dict.

        Parameters:
            prefix (str): Usually 'Initialized', 'Loaded' or 'Optimized' to identify the state of the hyperparameters.
        """

        print('\n'.join(["{} hyperparameters:".format(prefix)] +
                        ["{k}: {v}".format(k=key, v=value) for key, value in self.hyperparameters.items()]))


@Surrogate.register('Custom')
class GPSurrogate(GaussianProcess):
    """Custom GP model made from scratch.

     Supports custom Python and Fortran kernels with analytic derivatives and advanced Active Learning.

    Attributes:
        hess_inv (ndarray): Inverse Hessian matrix which is required for active learning.
            It is calculated during the hyperparameter optimization.
    """

    from .backend import gp_functions

    def __init__(self):
        super().__init__()
        self.hess_inv = None

    @property
    def Ky(self):
        """Full training covariance matrix as defined in the kernel
        including data noise as specified in the hyperparameters.
        """
        return self.kernel(self.Xtrain, self.Xtrain, **self.hyperparameters)

    @property
    def alpha(self):
        r"""Convenient matrix-vector product of the inverse training matrix and the training output data.
        The equation is solved either exactly or with a least squares approximation.

        $$
        \begin{equation}
        \alpha = K_y^{-1} y_{train}
        \end{equation}
        $$
        """
        try:
            return np.linalg.solve(self.Ky, self.ytrain)
        except np.linalg.LinAlgError:
            return np.linalg.lstsq(self.Ky, self.ytrain, rcond=1e-15)[0]

    def train(self, X, y, kernel=None, hyperparameters=None, fixed_sigma_n=False,
              eval_gradient=True, return_hess_inv=False, multi_output=False):
        """After initializing the model with a kernel function and initial hyperparameters,
        it can be trained on input data X and observed output data y by optimizing the model's hyperparameters.
        This is done by minimizing the negative log likelihood.

        Parameters:
            X (ndarray): (n, D) array of input training data.
            y (ndarray): (n, 1) array of training output.
                Currently, only scalar output is supported.
            kernel (str/object): Identifier of kernel like 'RBF' or directly the kernel object of the
                specific surrogate.
            hyperparameters (dict): Hyperparameters such as length_scale, variance and noise.
                Taken either from given parameter, config file or inferred from the training data.
                The hyperparameters can be different depending on the kernel. E.g. The length_scale can be a scalar,
                a vector of the size of the training data, or for the custom LinearEmbedding kernel a matrix.
            fixed_sigma_n (bool): Indicates if the data noise should be optimized or not.
            eval_gradient (bool): Whether the gradients of the kernel and negative log likelihood are
                explicitly used in the scipy optimization or numerically calculated inside scipy.
            return_hess_inv (bool): Whether to the attribute hess_inv after optimization. This is important
                for active learning.
            multi_output (bool): Indicates if a multi output is desired if y has more than one dimension.
                Otherwise the excess dimensions are used as independent variable.
        """

        super().prepare_train(X, y, kernel, hyperparameters, fixed_sigma_n)
        if self.encoder:
            print("For now, encoding is not supported for this surrogate. The model is created without encoding.")
            self.decode_training_data()

        if self.multi_output:
            raise NotImplementedError("Multi-Output is not implemented for this surrogate.")

        # Find best hyperparameters
        self.optimize(fixed_sigma_n=self.fixed_sigma_n, eval_gradient=eval_gradient, return_hess_inv=return_hess_inv)
        self.trained = True

    def predict(self, Xpred, add_data_variance=True):
        Xpred = super().prepare_predict(Xpred)
        # Encoding is not supported yet for this surrogate.
        for enc in self.encoder[::-1]:
            if not enc.output:
                Xpred = enc.decode(Xpred)

        # Skip data noise sigma_n in hyperparameters
        prediction_hyperparameters = {key: value for key, value in self.hyperparameters.items() if key != 'sigma_n'}

        # Calculate conditional mean and covariance functions
        Kstar = self.kernel(self.Xtrain, Xpred, **prediction_hyperparameters)
        Kstarstar_diag = np.diag(self.kernel(Xpred, Xpred, **prediction_hyperparameters))
        fstar = Kstar.T @ self.alpha
        vstar = Kstarstar_diag - np.diag((Kstar.T @ (self.gp_functions.invert(self.Ky) @ Kstar)))
        vstar = np.maximum(vstar, 1e-10)  # Assure a positive variance
        if add_data_variance:
            vstar = vstar + self.hyperparameters['sigma_n']**2
        return fstar, vstar.reshape(-1, 1)  # Return predictive mean and variance

    def add_training_data(self, X, y):
        """Add training points to existing data. This is important for active learning.

        Only the training dataset is updated, but the hyperparameters are not optimized yet.

        Parameters:
            X (ndarray): Input points to add.
            y (ndarray): Observed output to add.
        """

        # TODO: Update Ky by applying the Sherman-Morrison-Woodbury formula?
        self.Xtrain = np.concatenate([self.Xtrain, X], axis=0)
        self.ytrain = np.concatenate([self.ytrain, y], axis=0)

    def get_marginal_variance(self, Xpred):
        r"""Calculates the marginal variance to infer the next point in active learning.

        The calculation follows Osborne (2012).

        Derivation of the marginal variance:

            $\tilde{V}$ ... Marginal covariance matrix
            $\hat{V}$ ... Predictive variance
            $\frac{dm}{d\theta}$ ... Derivative of the predictive mean w.r.t. the hyperparameters
            $H$ ... Hessian matrix

            $$
            \begin{equation}
            \tilde{V} = \left( \frac{dm}{d\theta} \right) H^{-1} \left( \frac{dm}{d\theta} \right)^T
            \end{equation}
            $$

        Parameters:
            Xpred (ndarray): Possible prediction input points.

        Returns:
            ndarray: Sum of the actual marginal variance and the predictive variance.
        """
        mtilde, vhat = self.predict(Xpred)
        return self.gp_functions.marginal_variance_BBQ(self.Xtrain, self.ytrain, Xpred,
                                                       self.kernel, self.hyperparameters, self.hess_inv,
                                                       self.fixed_sigma_n, self.alpha, vhat)

    def save_model(self, path):
        """Saves the model as dict to a .hdf5 file.

        Parameters:
            path (str): Path including the file name, where the model should be saved.
        """
        from profit.util import save_hdf
        save_dict = {attr: getattr(self, attr)
                     for attr in ('trained', 'fixed_sigma_n', 'Xtrain', 'ytrain', 'ndim', 'kernel', 'hyperparameters')}

        # Convert the kernel class object to a string, to be able to save it in the .hdf5 file
        if not isinstance(save_dict['kernel'], str):
            save_dict['kernel'] = self.kernel.__name__
        save_hdf(path, save_dict)

    @classmethod
    def load_model(cls, path):
        """Loads a saved model from a .hdf5 file and updates its attributes.

        Parameters:
            path (str): Path including the file name, from where the model should be loaded.

        Returns:
            profit.sur.gaussian_process.GPSurrogate: Instantiated surrogate model.
        """

        from profit.util import load_hdf

        sur_dict = load_hdf(open(path, 'rb'), astype='dict')
        self = cls()

        for attr, value in sur_dict.items():
            setattr(self, attr, value)
        self.kernel = self.select_kernel(self.kernel)  # Convert the kernel string back to the class object
        self.print_hyperparameters("Loaded")
        return self

    def select_kernel(self, kernel):
        """Convert the name of the kernel as string to the kernel class object of the surrogate.
        First search the kernels implemented in python, then the Fortran kernels.

        Parameters:
            kernel (str): Kernel string such as 'RBF'. Only single kernels are supported currently.

        Returns:
            object: Kernel object of the class. This is the function which builds the kernel and not
            the calculated covariance matrix.
        """

        # TODO: Rewrite fortran kernels and rename them to be explicit, like 'fRBF'
        from .backend import kernels as fortran_kernels
        from .backend import python_kernels

        try:
            return getattr(python_kernels, kernel)
        except AttributeError:
            try:
                return getattr(fortran_kernels, kernel)
            except AttributeError:
                raise RuntimeError("Kernel {} not implemented.".format(kernel))

    def optimize(self, fixed_sigma_n=False, eval_gradient=False, return_hess_inv=False):
        r"""Optimize the hyperparameters length_scale $l$, scale $\sigma_f$ and noise $\sigma_n$.

        As a backend, the scipy minimize optimizer is used.

        Parameters:
            fixed_sigma_n (bool): Indication if the data noise should also be optimized or not.
            eval_gradient (bool): Flag if the gradients of the kernel and negative log likelihood should be
                used explicitly or numerically calculated inside the optimizer.
            return_hess_inv (bool): Whether to set the inverse Hessian attribute hess_inv which is used to calculate the
                marginal variance in active learning.
        """
        ordered_hyp_keys = ('length_scale', 'sigma_f', 'sigma_n')
        a0 = np.concatenate([self.hyperparameters[key] for key in ordered_hyp_keys])
        opt_hyperparameters = self.gp_functions.optimize(self.Xtrain, self.ytrain, a0, self.kernel,
                                                         fixed_sigma_n=self.fixed_sigma_n or fixed_sigma_n,
                                                         eval_gradient=eval_gradient, return_hess_inv=return_hess_inv)
        if return_hess_inv:
            self.hess_inv = opt_hyperparameters[1].todense()
            opt_hyperparameters = opt_hyperparameters[0]

        # Set optimized hyperparameters
        last_idx = -1 if self.fixed_sigma_n else -2
        self.hyperparameters['length_scale'] = np.atleast_1d(opt_hyperparameters[:last_idx])
        self.hyperparameters['sigma_f'] = np.atleast_1d(opt_hyperparameters[last_idx])
        if not self.fixed_sigma_n:
            self.hyperparameters['sigma_n'] = np.atleast_1d(opt_hyperparameters[-1])
        self.print_hyperparameters("Optimized")


@Surrogate.register('GPy')
class GPySurrogate(GaussianProcess):
    """Surrogate for https://github.com/SheffieldML/GPy.

    Attributes:
        model (GPy.models): Model object of GPy.
    """

    def __init__(self):
        import GPy
        self.GPy = GPy
        super().__init__()
        self.model = None

    def train(self, X, y, kernel=None, hyperparameters=None, fixed_sigma_n=False, return_hess_inv=False,
              multi_output=False):
        super().prepare_train(X, y, kernel, hyperparameters, fixed_sigma_n, multi_output)

        if not self.multi_output or self.output_ndim < 2:
            self.model = self.GPy.models.GPRegression(self.Xtrain, self.ytrain, self.kernel,
                                                      noise_var=self.hyperparameters['sigma_n'] ** 2)
        else:
            print("{}-D output detected. Using Coregionalization.".format(self.output_ndim))
            icm = self.GPy.util.multioutput.ICM(input_dim=self.ndim, num_outputs=self.output_ndim, kernel=self.kernel)
            self.model = self.GPy.models.GPCoregionalizedRegression(self.output_ndim * [self.Xtrain],
                                                                    [self.ytrain[:, d].reshape(-1, 1)
                                                                    for d in range(self.output_ndim)],
                                                                    kernel=icm)

        self.optimize()
        self._set_hyperparameters_from_model()
        self.print_hyperparameters("Optimized")
        self.trained = True
        self.decode_training_data()

    def add_training_data(self, X, y):
        """Adds training points to the existing dataset.

        This is important for Active Learning. The data is added but the hyperparameters are not optimized yet.

        Parameters:
            X (ndarray): Input points to add.
            y (ndarray): Observed output to add.
        """
        self.Xtrain = np.concatenate([self.Xtrain, X], axis=0)
        self.ytrain = np.concatenate([self.ytrain, y], axis=0)
        self.model.set_XY(self.Xtrain, self.ytrain)

    def predict(self, Xpred, add_data_variance=True):
        Xpred = super().prepare_predict(Xpred)
        if not self.multi_output:
            ymean, yvar = self.model.predict(Xpred, include_likelihood=add_data_variance)
        else:
            ymean = np.empty((Xpred.shape[0], self.output_ndim))
            yvar = ymean.copy()
            for d in range(self.output_ndim):
                newX = np.hstack([Xpred, np.ones((Xpred.shape[0], 1)) * d])
                noise_dict = {'output_index': newX[:, self.ndim:].astype(int)}
                ym, yv = self.model.predict(newX, Y_metadata=noise_dict)
                ymean[:, d], yvar[:, d] = ym.flatten(), yv.flatten()
        ymean, yvar = self.decode_predict_data(ymean, yvar)
        return ymean, yvar

    def get_marginal_variance(self, Xpred):
        """Calculates the marginal variance to infer the next point in active learning.

        Currently only the predictive variance is taken into account.

        Parameters:
            Xpred (ndarray): Possible prediction input points.

        Returns:
            ndarray: Currently only the predictive variance.
        """

        mtilde, vhat = self.predict(Xpred)
        return vhat.reshape(-1, 1)

    def save_model(self, path):
        """Save the model as dict to a .hdf5 file. GPy does not support to_dict method for Coregionalization
        (multi-output) models yet. It is then saved as a pickle file.

        Parameters:
            path (str): Path including the file name, where the model should be saved.
        """

        from profit.util import save_hdf

        try:
            sur_dict = {attr: getattr(self, attr) for attr in ('Xtrain', 'ytrain')}
            sur_dict['model'] = self.model.to_dict()
            sur_dict['encoder'] = str([enc.repr for enc in self.encoder])
            save_hdf(path, sur_dict)
        except NotImplementedError:
            # GPy does not support to_dict method for Coregionalization kernel yet.
            from pickle import dump
            from os.path import splitext
            print("Saving to .hdf5 not implemented yet for multi-output models. Saving to .pkl file instead.")
            dump([self.model, self.Xtrain, self.ytrain, str([enc.repr for enc in self.encoder])],
                 open(splitext(path)[0] + '.pkl', 'wb'))

    @classmethod
    def load_model(cls, path):
        """Loads a saved model from a .hdf5 file and updates its attributes. In case of a multi-output model, the .pkl
        file is loaded, since .hdf5 is not supported yet.

        Parameters:
            path (str): Path including the file name, from where the model should be loaded.

        Returns:
            GPy.models: Instantiated surrogate model.
        """

        from profit.util import load
        from .encoders import Encoder
        from GPy import models

        self = cls()
        try:
            sur_dict = load(path, as_type='dict')
            self.model = models.GPRegression.from_dict(sur_dict['model'])
            self.Xtrain = sur_dict['Xtrain']
            self.ytrain = sur_dict['ytrain']
            self.encoder = [Encoder(func, cols, out) for func, cols, out in eval(sur_dict['encoder'])]
        except (OSError, FileNotFoundError):
            from pickle import load as pload
            from os.path import splitext
            # Load multi-output model from pickle file
            print("File {} not found. Trying to find a .pkl file with multi-output instead.".format(path))
            self.model, self.Xtrain, self.ytrain, encoder_str = pload(open(splitext(path)[0] + '.pkl', 'rb'))
            self.encoder = [Encoder(func, cols, out) for func, cols, out in eval(encoder_str)]
            self.output_ndim = int(max(self.model.X[:, -1])) + 1
            self.multi_output = True

        # Initialize the encoder by encoding and decoding the training data once.
        self.encode_training_data()
        self.decode_training_data()

        self.kernel = self.model.kern
        self._set_hyperparameters_from_model()
        self.ndim = self.Xtrain.shape[-1]
        self.trained = True
        self.print_hyperparameters("Loaded")
        return self

    def select_kernel(self, kernel):
        """Get the GPy.kern kernel by matching the given string kernel identifier.

        Parameters:
            kernel (str): Kernel string such as 'RBF' or depending on the surrogate also product and sum kernels
                such as 'RBF+Matern52'.

        Returns:
            GPy.kern: GPy kernel object. Currently, for sum and product kernels, the initial hyperparameters are the
            same for all kernels.
        """

        try:
            if not any(operator in kernel for operator in ('+', '*')):
                return getattr(self.GPy.kern, kernel)(self.ndim,
                                                      lengthscale=self.hyperparameters.get('length_scale', [1]),
                                                      variance=self.hyperparameters.get('sigma_f', 1)**2,
                                                      ARD=len(self.hyperparameters.get('length_scale', [1])) > 1)
            else:
                from re import split
                full_str = split('([+*])', kernel)
                kern = []
                for key in full_str:
                    kern += [key if key in ('+', '*') else
                               'self.GPy.kern.{}({}, lengthscale={}, variance={})'
                                   .format(key, self.ndim,
                                           self.hyperparameters.get('length_scale', [1]),
                                           self.hyperparameters.get('sigma_f', 1)**2)]
                return eval(''.join(kern))
        except AttributeError:
            raise RuntimeError("Kernel {} is not implemented.".format(kernel))

    def optimize(self, return_hess_inv=False, **opt_kwargs):
        """For hyperparameter optimization the GPy base optimization is used.

        Currently, the inverse Hessian can not be retrieved, which limits the active learning effectivity.

        Parameters:
            return_hess_inv (bool): Is not considered currently.
            opt_kwargs: Keyword arguments used directly in the GPy base optimization.
        """

        self.model.optimize(**opt_kwargs)
        self._set_hyperparameters_from_model()

    def _set_hyperparameters_from_model(self):
        r"""Helper function to set the hyperparameter dict from the model.

        It depends on whether it is a single kernel or a combined one.
        """
        if hasattr(self.model.kern, 'lengthscale'):
            self.hyperparameters['length_scale'] = self.model.kern.lengthscale.values
            self.hyperparameters['sigma_f'] = np.sqrt(self.model.kern.variance)
            self.hyperparameters['sigma_n'] = np.sqrt(self.model.likelihood.variance)
        elif hasattr(self.model.kern, 'parts'):
            for part in self.model.kern.parts:
                for key, value in zip(part.parameter_names(), part.param_array):
                    value = np.atleast_1d(value)
                    if key == 'lengthscale':
                        self.hyperparameters['length_scale'] = value
                    elif key == 'variance':
                        self.hyperparameters['sigma_f'] = np.sqrt(value)
                    else:
                        self.hyperparameters[key] = value
            noise_var = self.model.likelihood.gaussian_variance(
                self.model.Y_metadata).reshape(-1, self.output_ndim, order='F')
            self.hyperparameters['sigma_n'] = np.sqrt(np.max(noise_var, axis=0))


@Surrogate.register('Sklearn')
class SklearnGPSurrogate(GaussianProcess):
    """Surrogate for https://github.com/scikit-learn/scikit-learn Gaussian process.

    Attributes:
        model (sklearn.gaussian_process.GaussianProcessRegressor): Model object of Sklearn.
    """

    def __init__(self):
        super().__init__()
        self.model = None

    def train(self, X, y, kernel=None, hyperparameters=None, fixed_sigma_n=False, return_hess_inv=False,
              multi_output=False):
        from sklearn.gaussian_process import GaussianProcessRegressor
        super().prepare_train(X, y, kernel, hyperparameters, fixed_sigma_n)

        if self.multi_output:
            raise NotImplementedError("Multi-Output is not implemented for this surrogate.")

        numeric_noise = self.hyperparameters['sigma_n'].item() ** 2 if self.fixed_sigma_n else 1e-5

        # Instantiate the model
        self.model = GaussianProcessRegressor(kernel=self.kernel, alpha=numeric_noise)

        # Train the model
        self.model.fit(self.Xtrain, self.ytrain)
        self.kernel = self.model.kernel_

        # Set hyperparameters from model
        self._set_hyperparameters_from_model()
        self.print_hyperparameters("Optimized")

        self.trained = True
        self.decode_training_data()

    def add_training_data(self, X, y):
        """Add training points to existing data.

        Parameters:
            X (ndarray): Input points to add.
            y (ndarray): Observed output to add.
        """
        self.Xtrain = np.concatenate([self.Xtrain, X], axis=0)
        self.ytrain = np.concatenate([self.ytrain, y], axis=0)

    def predict(self, Xpred, add_data_variance=True):
        Xpred = super().prepare_predict(Xpred)

        ymean, ystd = self.model.predict(Xpred, return_std=True)
        yvar = ystd.reshape(-1, 1)**2
        if add_data_variance:
            yvar = yvar + self.hyperparameters['sigma_n'] ** 2
        ymean, yvar = self.decode_predict_data(ymean, yvar)
        return ymean, yvar

    def get_marginal_variance(self, Xpred):
        """Calculates the marginal variance to infer the next point in active learning.

        Currently only the predictive variance is taken into account.

        Parameters:
            Xpred (ndarray): Possible prediction input points.
        Returns:
            ndarray: Currently only predictive variance.
        """

        mtilde, vhat = self.predict(Xpred)
        return vhat.reshape(-1, 1)

    def save_model(self, path):
        """Save the SklGPSurrogate model to a pickle file. All attributes of the surrogate are loaded directly from the
        model.

        Parameters:
            path (str): Path including the file name, where the model should be saved.
        """

        from pickle import dump
        dump(self.model, open(path, 'wb'))

    @classmethod
    def load_model(cls, path):
        """Load a saved SklGPSurrogate model from a pickle file and update its attributes.

        Parameters:
            path (str): Path including the file name, from where the model should be loaded.

        Returns:
            profit.sur.gaussian_process.SklearnGPSurrogate: Instantiated surrogate model.
        """

        from pickle import load

        self = cls()
        self.model = load(open(path, 'rb'))
        self.Xtrain = self.model.X_train_
        self.ytrain = self.model.y_train_
        self.kernel = self.model.kernel_
        self.ndim = self.Xtrain.shape[-1]
        self.fixed_sigma_n = self.model.alpha != 1e-5
        self.trained = True
        self._set_hyperparameters_from_model()
        self.print_hyperparameters("Loaded")
        return self

    def optimize(self, return_hess_inv=False, **opt_kwargs):
        """For hyperparameter optimization the Sklearn base optimization is used.

        Currently, the inverse Hessian can not be retrieved, which limits the active learning effectivity.

        Parameters:
            return_hess_inv (bool): Is not considered currently.
            opt_kwargs: Keyword arguments used directly in the Sklearn base optimization.
        """
        self.model.fit(self.Xtrain, self.ytrain, **opt_kwargs)
        self._set_hyperparameters_from_model()

    def select_kernel(self, kernel):
        """Get the sklearn.gaussian_process.kernels kernel by matching the given kernel identifier.

        Parameters:
            kernel (str): Kernel string such as 'RBF' or depending on the surrogate also product and sum kernels
                such as 'RBF+Matern52'.

        Returns:
            sklearn.gaussian_process.kernels: Scikit-learn kernel object. Currently, for sum and product kernels,
            the initial hyperparameters are the same for all kernels.
        """

        from re import split
        from sklearn.gaussian_process import kernels as sklearn_kernels
        full_str = split('([+*])', kernel)
        try:
            kernel = []
            for key in full_str:
                kernel += [key if key in ('+', '*') else
                           getattr(sklearn_kernels, key)(length_scale=self.hyperparameters['length_scale'])]
        except AttributeError:
            raise RuntimeError("Kernel {} is not implemented.".format(kernel))

        if len(kernel) == 1:
            kernel = kernel[0]
        else:
            kernel = [str(key) if not isinstance(key, str) else key for key in kernel]
            kernel = eval(''.join(kernel))

        # Add scale and noise to kernel
        kernel *= sklearn_kernels.ConstantKernel(constant_value=1/self.hyperparameters['sigma_f'].item() ** 2)
        if not self.fixed_sigma_n:
            kernel += sklearn_kernels.WhiteKernel(noise_level=self.hyperparameters['sigma_n'].item() ** 2)

        return kernel

    def _set_hyperparameters_from_model(self):
        r"""Helper function to set the hyperparameter dict from the model.

        It depends on whether $\sigma_n$ is fixed.
        Currently this is only stable for single kernels and not for Sum and Prod kernels.
        """
        if self.fixed_sigma_n:
            self.hyperparameters['length_scale'] = np.atleast_1d(self.model.kernel_.k1.length_scale)
            self.hyperparameters['sigma_f'] = np.sqrt(np.atleast_1d(1 / self.model.kernel_.k2.constant_value))
            self.hyperparameters['sigma_n'] = np.sqrt(np.atleast_1d(self.model.alpha))
        else:
            self.hyperparameters['length_scale'] = np.atleast_1d(self.model.kernel_.k1.k1.length_scale)
            self.hyperparameters['sigma_f'] = np.sqrt(np.atleast_1d(1 / self.model.kernel_.k1.k2.constant_value))
            self.hyperparameters['sigma_n'] = np.sqrt(np.atleast_1d(self.model.kernel_.k2.noise_level))



# Draft for Scikit-learn implementation of LinearEmbedding kernel of Garnett (2014)
from sklearn.gaussian_process.kernels import Kernel, Hyperparameter, StationaryKernelMixin, NormalizedKernelMixin
class LinearEmbedding(StationaryKernelMixin, NormalizedKernelMixin, Kernel):

    def __init__(self, dims, length_scale=np.array([1.0]), length_scale_bounds=(1e-5, 1e5)):
        self.length_scale = length_scale
        self.dims = dims
        self.length_scale_bounds = length_scale_bounds

    @property
    def hyperparameter_length_scale(self):
        return Hyperparameter("length_scale", "numeric",
                              self.length_scale_bounds,
                              len(self.length_scale))

    def __call__(self, X, Y=None, eval_gradient=False):
        X = np.atleast_2d(X)
        R = self.length_scale.reshape(self.dims)
        X1 = X @ R.T
        X2 = X @ R.T if Y is None else Y @ R.T
        dX = X1[:, np.newaxis, :] - X2[np.newaxis, :, :]
        dX_sq = np.linalg.norm(dX, axis=-1) ** 2
        K = np.exp(-0.5 * dX_sq)
        if eval_gradient:
            K_gradient = np.einsum('ijk,kl', dX ** 2, R) * K[..., np.newaxis]
            return K, K_gradient
        return K

    def __repr__(self):
        return "{0}(length_scale=[{1}])".format(self.__class__.__name__, ", ".join(map("{0:.3g}".format,
                                                                                       self.length_scale)))



""" mockup 1D example with a custom postprocessor """
from profit.run import Postprocessor, Worker

import numpy as np


@Postprocessor.register('mockup_post1')
class MockupPostprocessor(Postprocessor):
    """ almost identical copy of NumpytxtPostprocessor """
    def post(self, data):
        raw = np.loadtxt('mockup.out')
        data['f'] = raw

    @classmethod
    def handle_config(cls, config, base_config):
        pass


@Postprocessor.wrap('mockup_post2')
def post(data):
    """ shorthand for mockup_post1 """
    raw = np.loadtxt('mockup.out')
    data['f'] = raw


@Worker.register('mockup_worker2')
class Mockup(Worker):
    """ directly calling the wanted python function """
    def main(self):
        u = self.interface.input['u']
        self.interface.output['f'] = np.cos(10 * u) + u
        self.interface.done()


@Worker.register('mockup_worker3')
class Mockup(Worker):
    """ substituting the run method to do something special, pre and post is executed as usual """
    def run(self):
        """ this is actually pretty stupid, as the inputs and outputs are also directly accessible via the interface """
        u = np.loadtxt('mockup_1D.in')
        f = np.cos(10 * u) + u
        print('custom3', f)
        np.savetxt('mockup.out', np.array([f]))


@Worker.wrap('mockup_worker4', 'u', 'f')
def mockup4(u):
    """ shorthand for mockup_worker2 """
    return np.cos(10 * u) + u




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Aug  9 22:30:35 2020

@author: manal
"""

def __bootstrap__():
   global __bootstrap__, __loader__, __file__
   import sys, pkg_resources, imp
   __file__ = pkg_resources.resource_filename(__name__,'kernels.cpython-38-x86_64-linux-gnu.so')
 #  __file__ = pkg_resources.resource_filename(__name__,'kernels.cpython-38-x86_64-linux-gnu.so')
   __loader__ = None; del __bootstrap__, __loader__
   imp.load_dynamic(__name__,__file__)
__bootstrap__()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Aug 22 00:21:20 2020
@author: manal
"""

import numpy as np
import GPy
from GPy.kern.src.stationary import Stationary
import matplotlib.pyplot as plt

class Cosin(Stationary):
    """
    Cosine kernel:
        
    .. math::
       k(r) = \sigma^2 \cos ( r )
       
       r(x, x') = \\sqrt{ \\sum_{q=1}^Q \\frac{(x_q - x'_q)^2}{\ell_q^2} }
    """
    def __init__(self, input_dim, variance=1., lengthscale=None, ARD=False, active_dims=None, name='Cosin'):
        super(Cosin, self).__init__(input_dim, variance, lengthscale, ARD, active_dims, name)

    def K_of_r(self, r):
        return self.variance * np.cos(r)

    def dK_dr(self, r):                     # returns the 1st derivative of the kernel wrt r
        return -self.variance * np.sin(r)
    
    def dK_dX(self, X, X2, dimX):           # returns the 1st derivative wrt X
        r = self._scaled_dist(X, X2)        # radial distance between X and X2
        dK_dr = self.dK_dr(r)               # 1st derivative wrt r
        dist = X[:,None,dimX]-X2[None,:,dimX]                                   # distance between X and X2
        lengthscale2inv = (np.ones((X.shape[1]))/(self.lengthscale**2))[dimX]   # 1/(lengthscale**2)
        return lengthscale2inv * dist * r**(-1) * dK_dr
    
    def dK_dX2(self,X,X2,dimX2):            # returns the 1st derivative wrt X2
        return -1.*self.dK_dX(X,X2, dimX2)
    
    def dK2_dXdX2(self, X, X2, dimX, dimX2):# returns the 2nd derivative wrt X and X2
        if X2 is None:
            X2 = X
        r = self._scaled_dist(X, X2)
        dist = X[:,None,:]-X2[None,:,:]
        lengthscale2inv = np.ones((X.shape[1]))/(self.lengthscale**2)
        l1 = lengthscale2inv[dimX]          # 1/(l_dimX)**2
        l2 = lengthscale2inv[dimX2]         # 1/(l_dimX2)**2
        d1 = dist[:,:,dimX]                 # (X_dimX - X2_dimX)
        d2 = dist[:,:,dimX2]                # (X_dimX2 - X2_dimX2)
        return (dimX!=dimX2)*self.variance*d1*d2*l1*l2*(r*np.cos(r)-np.sin(r))*r**(-3) + (dimX==dimX2)*self.variance*l2*(r**2*np.sin(r)+r*d1**2*l2*np.cos(r)-d1**2*l2*np.sin(r))*r**(-3)



"""
Created: Tue Mar 26 09:45:46 2019
@author: Christopher Albert <albert@alumni.tugraz.at>
"""
from profit import uq
import numpy as np
from fffi import FortranModule

from time import time

mean1 = 5.0
std1 = 0.1
mean2 = 2.0
std2 = 0.2

uqp = FortranModule('uqp', 'mod_unqu')

uqp.fdef("""
  integer :: nporder, nall, npar, nt, iflag_run, iflag_mod, iflag_pol

  subroutine allocate_params end
  subroutine init_uq end
  
  subroutine set_legendre_borders(kpar, lower, upper)
    integer, intent(in) :: kpar
    double precision, intent(in) :: lower, upper
  end

  subroutine set_hermite_mean_std(kpar, mean0, std)
    integer, intent(in) :: kpar
    double precision, intent(in) :: mean0, std
  end
  
  subroutine pre_uq(axi)
    double precision, intent(inout) :: axi(:,:)
  end
  
  subroutine run_uq end
  """)

uqp.compile(verbose=1)
uqp.load()
#%%
print('=== Hermite 2D quadrature points ===')

t = time()  # UQP timing
uqp.npar = 2
uqp.allocate_params()

uqp.nporder = 30
uqp.nt = 1
uqp.iflag_run = 1
uqp.iflag_mod = 2
uqp.iflag_pol = 2  # 1: Legende, 2: Hermite, 3: Laguerre
uqp.set_hermite_mean_std(1, mean1, std1)
uqp.set_hermite_mean_std(2, mean2, std2)
uqp.init_uq()

axi = np.zeros((uqp.nall,uqp.npar), order='F')
uqp.pre_uq(axi)

print('UQP: {:4.2f} ms'.format(1000*(time() - t)))

t = time()  # ChaosPy timing
uq.backend = uq.ChaosPy(order = uqp.nporder, sparse = False)
uq.params['u'] = uq.Normal(mean1, std1)
uq.params['v'] = uq.Normal(mean2, std2)
axi2 = uq.get_eval_points().T

print('ChaosPy: {:4.2f} ms'.format(1000*(time() - t)))

axis = axi[np.lexsort((axi[:,0], axi[:,1]))]
axi2s = axi2[np.lexsort((axi2[:,0], axi2[:,1]))]

np.testing.assert_almost_equal(axis, axi2s)
print('Success: UQP and ChaosPy results match')

# TODO: move this into unit test



""" Default Runner & Worker components

Local Runner
Memmap Interface (numpy)
Template Preprocessor
JSON Postprocessor
NumpytxtPostprocessor
HDF5Postprocessor
"""

from .runner import Runner, RunnerInterface
from .worker import Interface, Preprocessor, Postprocessor, Worker

import subprocess
from multiprocessing import Process
from time import sleep
import logging

import numpy as np
import os
from shutil import rmtree

# === Local Runner === #


@Runner.register('local')
class LocalRunner(Runner):
    """ Runner for executing simulations locally

    - forks the worker, thereby having less overhead (especially with a custom python Worker)
    - per default uses all available CPUs
    """
    def spawn_run(self, params=None, wait=False):
        super().spawn_run(params, wait)
        if self.run_config['custom'] or not self.config['fork']:
            env = self.env.copy()
            env['PROFIT_RUN_ID'] = str(self.next_run_id)
            if self.run_config['custom']:
                cmd = self.run_config['command']
            else:
                cmd = 'profit-worker'
            self.runs[self.next_run_id] = subprocess.Popen(cmd, shell=True, env=env, cwd=self.base_config['run_dir'])
            if wait:
                self.runs[self.next_run_id].wait()
                del self.runs[self.next_run_id]
        else:
            os.chdir(self.base_config['run_dir'])
            worker = Worker.from_config(self.run_config, self.next_run_id)
            process = Process(target=worker.main)
            self.runs[self.next_run_id] = (worker, process)
            process.start()
            if wait:
                process.join()
                del self.runs[self.next_run_id]
            os.chdir(self.base_config['base_dir'])
        self.next_run_id += 1

    def spawn_array(self, params_array, blocking=True):
        """ spawn an array of runs, maximum 'parallel' at the same time, blocking until all are done """
        if not blocking:
            raise NotImplementedError
        for params in params_array:
            self.spawn_run(params)
            while len(self.runs) >= self.config['parallel']:
                sleep(self.config['sleep'])
                self.check_runs(poll=True)
        while len(self.runs):
            sleep(self.config['sleep'])
            self.check_runs(poll=True)

    def check_runs(self, poll=False):
        """ check the status of runs via the interface """
        self.interface.poll()
        if self.run_config['custom'] or not self.config['fork']:
            for run_id, process in list(self.runs.items()):  # preserve state before deletions
                if self.interface.internal['DONE'][run_id]:
                    process.wait()  # just to make sure
                    del self.runs[run_id]
                elif poll and process.poll() is not None:
                    del self.runs[run_id]
        else:
            for run_id, (worker, process) in list(self.runs.items()):  # preserve state before deletions
                if self.interface.internal['DONE'][run_id]:
                    process.join()  # just to make sure
                    del self.runs[run_id]
                elif poll and process.exitcode is not None:
                    process.terminate()
                    del self.runs[run_id]

    def cancel_all(self):
        if self.run_config['custom'] or not self.config['fork']:
            for process in self.runs.values():
                process.terminate()
        else:
            for worker, process in self.runs.values():
                process.terminate()
        self.runs = {}

    @classmethod
    def handle_config(cls, config, base_config):
        """
        Example:
            .. code-block:: yaml

                class: local
                parallel: all   # maximum number of simultaneous runs (for spawn array)
                sleep: 0        # number of seconds to sleep while polling
                fork: true      # whether to spawn the worker via forking instead of a subprocess (via a shell)
        """
        if 'parallel' not in config or config['parallel'] == 'all':
            config['parallel'] = os.cpu_count()
        if 'sleep' not in config:
            config['sleep'] = 0
        if 'fork' not in config:
            config['fork'] = True


# === Numpy Memmap Inerface === #


@RunnerInterface.register('memmap')
class MemmapRunnerInterface(RunnerInterface):
    """ Runner-Worker Interface using a memory mapped numpy array

    - expected to be very fast with the *local* Runner as each Worker can access the array directly (unverified)
    - expected to be inefficient if used on a cluster with a shared filesystem (unverified)
    - reliable
    - known issue: resizing the array (to add more runs) is dangerous, needs a workaround
      (e.g. several arrays in the same file)
    """
    def __init__(self, config, size, input_config, output_config, *, logger_parent: logging.Logger = None):
        super().__init__(config, size, input_config, output_config, logger_parent=logger_parent)

        init_data = np.zeros(size, dtype=self.input_vars + self.internal_vars + self.output_vars)
        np.save(self.config['path'], init_data)

        try:
            self._memmap = np.load(self.config['path'], mmap_mode='r+')
        except FileNotFoundError:
            self.runner.logger.error(
                f'{self.__class__.__name__} could not load {self.config["path"]} (cwd: {os.getcwd()})')
            raise

        # should return views on memmap
        self.input = self._memmap[[v[0] for v in self.input_vars]]
        self.output = self._memmap[[v[0] for v in self.output_vars]]
        self.internal = self._memmap[[v[0] for v in self.internal_vars]]

    def clean(self):
        if os.path.exists(self.config['path']):
            os.remove(self.config['path'])

    @classmethod
    def handle_config(cls, config, base_config):
        """
        Example:
            .. code-block:: yaml

                class: memmap
                path: interface.npy     # path to memory mapped interface file, relative to base directory
        """
        if 'path' not in config:
            config['path'] = 'interface.npy'
        # 'path' is relative to base_dir, convert to absolute path
        if not os.path.isabs(config['path']):
            config['path'] = os.path.abspath(os.path.join(base_config['base_dir'], config['path']))


@Interface.register('memmap')
class MemmapInterface(Interface):
    """ Runner-Worker Interface using a memory mapped numpy array

    counterpart to :py:class:`MemmapRunnerInterface`
    """
    def __init__(self, config, run_id: int, *, logger_parent: logging.Logger = None):
        super().__init__(config, run_id, logger_parent=logger_parent)
        # ToDo: multiple arrays after another to allow extending the file dynamically
        try:
            self._memmap = np.load(self.config['path'], mmap_mode='r+')
        except FileNotFoundError:
            self.worker.logger.error(
                f'{self.__class__.__name__} could not load {self.config["path"]} (cwd: {os.getcwd()})')
            raise

        # should return views on memmap
        inputs, outputs = [], []
        k = 0
        for k, key in enumerate(self._memmap.dtype.names):
            if key == 'DONE':
                break
            inputs.append(key)
        for key in self._memmap.dtype.names[k:]:
            if key not in ['DONE', 'TIME']:
                outputs.append(key)
        self.input = self._memmap[inputs][run_id]
        self.output = self._memmap[outputs][run_id]
        self._data = self._memmap[run_id]

    def done(self):
        self._memmap['TIME'] = self.time
        self._memmap['DONE'] = True
        self._memmap.flush()

    def clean(self):
        if os.path.exists(self.config['path']):
            os.remove(self.config['path'])


# === Template Preprocessor === #


@Preprocessor.register('template')
class TemplatePreprocessor(Preprocessor):
    """ Preprocessor which substitutes the variables with a given template

    - copies the given template directory to the target run directory
    - searches all files for variables templates of the form {name} and replaces them with their values
    - for file formats which use curly braces (e.g. json) the template identifier is {{name}}
    - substitution can be restricted to certain files by specifying `param_files`
    - relative symbolic links are converted to absolute symbolic links on copying
    - linked files are ignored with `param_files: all`, but if specified explicitly the link target is copied to the run
      directory and then substituted
    """
    def pre(self, data, run_dir):
        # No call to super()! replaces the default preprocessing
        from profit.pre import fill_run_dir_single
        if os.path.exists(run_dir):
            rmtree(run_dir)
        fill_run_dir_single(data, self.config['path'], run_dir, ignore_path_exists=True,
                            param_files=self.config['param_files'])
        os.chdir(run_dir)

    @classmethod
    def handle_config(cls, config, base_config):
        """
        Example:
            .. code-block:: yaml

                class: template
                path: template      # directory to copy from, relative to base directory
                param_files: null   # files in template which contain placeholders for variables, null means all files
                                    # can be a filename or a list of filenames
        """
        if 'path' not in config:
            config['path'] = 'template'
        # 'path' is relative to base_dir, convert to absolute path
        if not os.path.isabs(config['path']):
            config['path'] = os.path.abspath(os.path.join(base_config['base_dir'], config['path']))
        if 'param_files' not in config:
            config['param_files'] = None
        if isinstance(config['param_files'], str):
            config['param_files'] = [config['param_files']]


# === JSON Postprocessor === #


@Postprocessor.register('json')
class JSONPostprocessor(Postprocessor):
    """ Postprocessor to read output from a JSON file

    - variables are assumed to be stored with the correct key and able to be converted immediately
    - not extensively tested
    """
    def post(self, data):
        import json
        with open(self.config['path']) as f:
            output = json.load(f)
        for key, value in output.items():
            data[key] = value

    @classmethod
    def handle_config(cls, config, base_config):
        """
        Example:
            .. code-block:: yaml

                class: json
                path: stdout    # file to read from, relative to the run directory
        """
        if 'path' not in config:
            config['path'] = 'stdout'


# === Numpy Text Postprocessor === #


@Postprocessor.register('numpytxt')
class NumpytxtPostprocessor(Postprocessor):
    """ Postprocessor to read output from a tabular text file (e.g. csv, tsv) with numpy ``genfromtxt``

    - the data is assumed to be row oriented
    - vector variables are spread across the row and have to be in the right order, only the name of the variable should
      be specified once in ``names``
    - ``names`` which are not specified as output variables are ignored
    - additional options are passed directly to ``numpy.genfromtxt()`
    """
    def post(self, data):
        dtype = [(name, float, data.dtype[name].shape if name in data.dtype.names else ())
                 for name in self.config['names']]
        try:
            raw = np.genfromtxt(self.config['path'], dtype=dtype, **self.config['options'])
        except OSError:
            self.logger.error(f'output file {self.config["path"]} not found')
            self.logger.info(f'cwd = {os.getcwd()}')
            dirname = os.path.dirname(self.config['path']) or '.'
            self.logger.info(f'ls {dirname} = {os.listdir(dirname)}')
            raise
        for key in self.config['names']:
            if key in data.dtype.names:
                data[key] = raw[key]

    @classmethod
    def handle_config(cls, config, base_config):
        """
        Example:
            .. code-block:: yaml

                class: numpytxt
                path: stdout    # file to read from, relative to the run directory
                names: "f g"    # list or string of output variables in order, default read from config/variables
                options:        # options which are passed on to numpy.genfromtxt() (fname & dtype are used internally)
                    deletechars: ""
        """
        if 'path' not in config:
            config['path'] = 'stdout'
        if 'names' not in config:
            config['names'] = list(base_config['output'].keys())
        if isinstance(config['names'], str):
            config['names'] = config['names'].split()
        if 'options' not in config:
            config['options'] = {}
        if 'deletechars' not in config['options']:
            config['options']['deletechars'] = ""


# === HDF5 Postprocessor === #


@Postprocessor.register('hdf5')
class HDF5Postprocessor(Postprocessor):
    """ Postprocessor to read output from a HDF5 file

        - variables are assumed to be stored with the correct key and able to be converted immediately
        - not extensively tested
    """
    def post(self, data):
        import h5py
        with h5py.File(self.config['path'], 'r') as f:
            for key in f.keys():
                data[key] = f[key]

    @classmethod
    def handle_config(cls, config, base_config):
        """
        Example:
            .. code-block:: yaml

                class: hdf5
                path: output.hdf5   # file to read from, relative to the run directory
        """
        if 'path' not in config:
            config['path'] = 'output.hdf5'



"""
Collection of functions for the Custom GPSurrogate.
"""

import numpy as np


def optimize(xtrain, ytrain, a0, kernel, fixed_sigma_n=False, eval_gradient=False, return_hess_inv=False):
    r"""Finds optimal hyperparameters from initial array a0, sorted as [length_scale, scale, noise].

    The objective function, which is minimized, is the negative log likelihood.
    The hyperparameters are transformed logarithmically before the optimization to enhance its flexibility. They are
    transformed back afterwards.

    Parameters:
        xtrain (ndarray):
        ytrain (ndarray):
        a0 (ndarray): Flat array of initial hyperparameters.
        kernel (function): Function of the kernel, not the actual matrix.
        fixed_sigma_n (bool): If the noise $\sigma_n$ should be fixed during optimization.
        eval_gradient (bool): Whether the analytic derivatives of the kernel and likelihood are used in the
            optimization.
        return_hess_inv (bool): If True, returns the inverse Hessian matrix from the optimization result.
            This is important for advanced active learning.

    Returns:
        tuple: A tuple containing:
            - opt_hyperparameters (ndarray): Flat array of optimized hyperparameters.
            - hess_inv (scipy.optimize.LbfgsInvHessProduct): Inverse Hessian matrix in the form of a scipy linear
            operator.
            If return_hess_inv is False, only the optimized hyperparameters are returned.
    """

    from scipy.optimize import minimize

    if fixed_sigma_n:
        sigma_n = a0[-1]
        a0 = a0[:-1]
    else:
        sigma_n = None

    a0 = np.log10(a0)
    bounds = [(None, None)] * len(a0)

    # Additional arguments for the negative_log_likelihood function.
    args = [xtrain, ytrain, kernel, eval_gradient, True]

    # If sigma_n should be kept fixed during optimization.
    if sigma_n is not None:
        args.append(sigma_n)

    opt_result = minimize(negative_log_likelihood, a0, args=tuple(args), bounds=bounds, jac=eval_gradient)
    if return_hess_inv:
        return 10**opt_result.x, opt_result.hess_inv
    return 10**opt_result.x


def solve_cholesky(L, b):
    r"""Solves a linear equation with a lower triangular matrix L from the Cholesky decomposition.

    In the context of GP's the $L$ is the Cholesky decomposition of the training kernel matrix and $b$ is the training
    output data.

    $$
    \begin{equation}
    \alpha = L^T L b
    \end{equation}
    $$

    Parameters:
        L (ndarray): Lower triangular matrix.
        b (ndarray): A vector.

    Returns:
        ndarray
    """

    from scipy.linalg import solve_triangular
    alpha = solve_triangular(L.T, solve_triangular(L, b, lower=True, check_finite=False),
                             lower=False, check_finite=False)
    return alpha


def negative_log_likelihood_cholesky(hyp, X, y, kernel, eval_gradient=False, log_scale_hyp=False, fixed_sigma_n=False):
    r"""Computes the negative log likelihood using the Cholesky decomposition of the covariance matrix.

    The calculation follows Rasmussen&Williams 2006, p. 19, 113-114.

    $$
    \begin{align}
    NL &= \frac{1}{2} y^T \alpha + tr(\log(L)) + \frac{n}{2} \log(2 \pi) \\
    \frac{dNL}{d\theta} &= \frac{1}{2} tr\left( (K_y^{-1} - \alpha \alpha^T) \frac{\partial K}{\partial \theta} \right) \\
    \alpha &= K_y^{-1} y
    \end{align}
    $$

    Parameters:
        hyp (ndarray): Flat hyperparameter array [length_scale, sigma_f, sigma_n]. They can also be already
            log transformed.
        X (ndarray): Training input points.
        y (ndarray): Observed training output.
        kernel (function): Function to build the covariance matrix.
        eval_gradient (bool): If the analytic gradient of the negative log likelihood w.r.t. the hyperparameters should
            be returned.
        log_scale_hyp (bool): Whether the hyperparameters are log transformed. This is important for the gradient
            calculation.
        fixed_sigma_n (bool): If the noise $\sigma_n$ is kept fixed. In this case, there is no gradient with respect to
            sigma_n.

    Returns:
        tuple: A tuple containing:
            - nll (float): The negative log likelihood.
            - dnll (ndarray): The derivative of the negative log likelihood w.r.t. to the hyperparameters.
            If eval_gradient is False, only nll is returned.
    """

    Ky = kernel(X, X, hyp[:-2], *hyp[-2:], eval_gradient=eval_gradient)
    if eval_gradient:
        dKy = Ky[1]
        Ky = Ky[0]

    L = np.linalg.cholesky(Ky)
    alpha = solve_cholesky(L, y)
    nll = 0.5 * y.T @ alpha + np.sum(np.log(L.diagonal())) + len(X) * 0.5 * np.log(2.0 * np.pi)
    if not eval_gradient:
        return nll.item()
    KyinvaaT = invert_cholesky(L)
    KyinvaaT -= np.outer(alpha, alpha)
    dnll = 0.5 * np.trace(KyinvaaT @ dKy)  # Rasmussen&Williams p. 114, eq. 5.9
    if fixed_sigma_n:
        dnll = dnll[:-1]
        hyp = hyp[:-1]
    if log_scale_hyp:
        # Chain rule
        dnll *= hyp * np.log(10)
    return nll.item(), dnll


def negative_log_likelihood(hyp, X, y, kernel, eval_gradient=False, log_scale_hyp=False,
                            fixed_sigma_n_value=None, neig=0, max_iter=1000):
    r"""Computes the negative log likelihood either by a Cholesky- or an Eigendecomposition.

    First, the Cholesky decomposition is tried. If this results in a LinAlgError, the biggest eigenvalues are
    calculated until convergence or until the maximum iterations are reached.
    The eigenvalues are cut off at $1e-10$ due ensure numerical stability.

    $$
    \begin{align}
    NL &= \frac{1}{2} \left( y^T \alpha + tr(\log(\lambda)) + n_{eig} \log(2 \pi) \right) \\
    \frac{dNL}{d\theta} &= \frac{1}{2} tr\left( (K_y^{-1} - \alpha \alpha^T) \frac{\partial K}{\partial \theta} \right) \\
    \alpha &= v (\lambda^{-1} (v^T y))
    \end{align}
    $$

    Parameters:
        hyp (ndarray): Flat hyperparameter array [length_scale, sigma_f, sigma_n].
        X (ndarray): Training input points.
        y (ndarray): Observed training output.
        kernel (function): Function to build the covariance matrix.
        eval_gradient (bool): If the analytic gradient of the negative log likelihood w.r.t. the hyperparameters should
            be returned.
        log_scale_hyp (bool): Whether the hyperparameters are log transformed. This is important for the gradient
            calculation.
        fixed_sigma_n_value (float): The value of the fixed noise $\sigma_n$. If it should be optimizied as well, this
            should be None.
        neig (int): Initial number of eigenvalues to calculate if the Cholesky decomposition is not successful.
            This is doubled during every iteration.
        max_iter (int): Maximum number of iterations of the eigenvalue solver until convergence must be reached.

    Returns:
        tuple: A tuple containing:
            - nll (float): The negative log likelihood.
            - dnll (ndarray): The derivative of the negative log likelihood w.r.t. to the hyperparameters.
            If eval_gradient is False, only nll is returned.
    """

    from scipy.sparse.linalg import eigsh

    if log_scale_hyp:
        hyp = 10**hyp
    fixed_sigma_n = fixed_sigma_n_value is not None
    if fixed_sigma_n:
        hyp = np.append(hyp, fixed_sigma_n_value)

    clip_eig = max(1e-3 * min(abs(hyp[:-2])), 1e-10)
    Ky = kernel(X, X, hyp[:-2], *hyp[-2:], eval_gradient=eval_gradient)  # Construct covariance matrix

    if eval_gradient:
        dKy = Ky[1]
        Ky = Ky[0]

    converged = False
    iteration = 0
    neig = max(neig, 1)
    while not converged:
        if not neig or neig > 0.05 * len(X):  # First try with Cholesky decomposition if neig is big
            try:
                return negative_log_likelihood_cholesky(hyp, X, y, kernel, eval_gradient=eval_gradient,
                                                        log_scale_hyp=log_scale_hyp, fixed_sigma_n=fixed_sigma_n)
            except np.linalg.LinAlgError:
                print("Warning! Fallback to eig solver!")
        w, Q = eigsh(Ky, neig, tol=clip_eig)  # Otherwise, calculate the first neig eigenvalues and eigenvectors
        if iteration >= max_iter:
            print("Reached max. iterations!")
            break
        neig *= 2  # Calculate more eigenvalues
        converged = w[0] <= clip_eig or neig >= len(X)

    # Calculate the NLL with these eigenvalues and eigenvectors
    w = np.maximum(w, 1e-10)
    alpha = Q @ (np.diag(1.0 / w) @ (Q.T @ y))
    nll = 0.5 * (y.T @ alpha + np.sum(np.log(w)) + min(neig, len(X)) * np.log(2 * np.pi))
    if not eval_gradient:
        return nll.item()

    KyinvaaT = invert(Ky)
    KyinvaaT -= np.outer(alpha, alpha)

    dnll = 0.5 * np.trace(KyinvaaT @ dKy)  # Rasmussen&Williams p. 114, eq. 5.9
    if fixed_sigma_n:
        dnll = dnll[:-1]
        hyp = hyp[:-1]
    if log_scale_hyp:
        # Chain rule
        dnll *= hyp * np.log(10)
    return nll.item(), dnll


def invert_cholesky(L):
    r"""Inverts a positive-definite matrix based on a Cholesky decomposition.

    This is used to invert the covariance matrix.

    Parameters:
        L (ndarray): Lower triangular matrix from a Cholesky decomposition.

    Returns:
        ndarray: Inverse of the matrix $L^T L$
    """
    from scipy.linalg import solve_triangular

    return solve_triangular(L.T, solve_triangular(L, np.eye(L.shape[0]), lower=True, check_finite=False),
                            lower=False, check_finite=False)


def invert(K, neig=0, tol=1e-10, max_iter=1000):
    """Inverts a positive-definite matrix using either a Cholesky- or an Eigendecomposition.

    The solution method depends on the rapidness of decay of the eigenvalues.

    Parameters:
        K (ndarray): Kernel matrix.
        neig (int): Initial number of eigenvalues to calculate if the Cholesky decomposition is not successful.
            This is doubled during every iteration.
        tol (float): Convergence criterion for the eigenvalues.
        max_iter (int): Maximum number of iterations of the eigenvalue solver until convergence must be reached.

    Returns:
        ndarray: Inverse covariance matrix.
    """

    # TODO: Fix this or continue using scipy?
    """
    L = np.linalg.cholesky(K)  # Cholesky decomposition of the covariance matrix
    if neig <= 0 or neig > 0.05 * len(K):  # First try with Cholesky decomposition
        try:
            return cls.invert_cholesky(L)
        except np.linalg.LinAlgError:
            print('Warning! Fallback to eig solver!')

    # Otherwise, calculate the first neig eigenvalues and eigenvectors
    w, Q = eigsh(K, neig, tol=tol)
    for iteration in range(max_iter):  # Iterate until convergence or max_iter
        if neig > 0.05 * len(K):
            try:
                return cls.invert_cholesky(L)
            except np.linalg.LinAlgError:
                print("Warning! Fallback to eig solver!")
        neig = 2 * neig  # Calculate more eigenvalues
        w, Q = eigsh(K, neig, tol=tol)

        # Convergence criterion
        if np.abs(w[0] - tol) <= tol or neig >= len(K):
            break
    if iteration == max_iter:
        print("Tried maximum number of times.")
    """
    from scipy.sparse.linalg import eigsh
    from scipy.linalg import eigh, inv
    # w, Q = eigh(K)
    # negative_eig = (-1e-2 < w) & (w < 0)
    # w[negative_eig] = 1e-2
    # K_inv = Q @ (np.diag(1 / w) @ Q.T)
    K_inv = inv(K, check_finite=True)
    return K_inv


def marginal_variance_BBQ(Xtrain, ytrain, Xpred, kernel, hyperparameters, hess_inv, fixed_sigma_n=False,
                          alpha=None, predictive_variance=0):
    r"""Calculates the marginal variance to infer the next point in active learning.

    The calculation follows Osborne (2012).

        $\tilde{V}$ ... Marginal covariance matrix
        $\hat{V}$ ... Predictive variance
        $\frac{dm}{d\theta}$ ... Derivative of the predictive mean w.r.t. the hyperparameters
        $H$ ... Hessian matrix

        $$
        \begin{equation}
        \tilde{V} = \left( \frac{dm}{d\theta} \right) H^{-1} \left( \frac{dm}{d\theta} \right)^T
        \end{equation}
        $$

    Parameters:
        Xtrain (ndarray): Input training points.
        ytrain (ndarray): Observed output daat.
        Xpred (ndarray): Possible prediction input points.
        kernel (function): Function to build the covariance matrix.
        hyperparameters (dict): Dictionary of the hyperparameters.
        hess_inv (ndarray): Inverse Hessian matrix.
        fixed_sigma_n (bool): If the noise $\sigma_n$ was fixed during optimization. Then, the Hessian has to be
            padded with zeros.
        alpha (ndarray): If available, the result of $K_y^{-1} y$, else None.
        predictive_variance (ndarray): Predictive variance only. This is added to the marginal variance.

    Returns:
        ndarray: Sum of the actual marginal variance and the predictive variance.
    """

    # TODO: Add full derivatives as in Osborne (2021) and Garnett (2014)
    ordered_hyperparameters = [hyperparameters[key] for key in ('length_scale', 'sigma_f', 'sigma_n')]

    # If no Hessian is available, use only the predictive variance.
    if hess_inv is None:
        return predictive_variance.reshape(-1, 1)

    if fixed_sigma_n:
        padding = np.zeros((len(ordered_hyperparameters), 1))
        hess_inv = np.hstack([np.vstack([hess_inv, padding[:-1].T]), padding])

    # Kernels and their derivatives
    Ky, dKy = kernel(Xtrain, Xtrain, *ordered_hyperparameters, eval_gradient=True)
    Kstar, dKstar = kernel(Xpred, Xtrain, *ordered_hyperparameters[:-1], eval_gradient=True)
    Kyinv = invert(Ky)
    if alpha is None:
        alpha = Kyinv @ ytrain
    dalpha_dl = -Kyinv @ (dKy[..., 0] @ alpha)
    dalpha_ds = -Kyinv @ (np.eye(Xtrain.shape[0]) @ alpha)

    # TODO: check derivatives
    dm = np.empty((Xpred.shape[0], len(ordered_hyperparameters), 1))
    dm[:, 0, :] = dKstar[..., 0] @ alpha - Kstar @ dalpha_dl
    dm[:, 1, :] = Kstar @ dalpha_ds
    dm[:, 2, :] = Kstar @ dalpha_ds
    dm = dm.squeeze()

    marginal_variance = dm @ (hess_inv @ dm.T)
    marginal_variance = np.diag(marginal_variance).reshape(-1, 1)

    return marginal_variance + predictive_variance


def predict_f(hyp, x, y, xtest, kernel, return_full_cov=False, neig=0):
    """Predicts values given only a set of hyperparameters and a kernel.

    This function is independent of the surrogates and used to quickly predict a function output with specific
    hyperparameters.

    The calculation follows Rasmussen&Williams, p. 16, eq. 23-24.

    Parameters:
        hyp (ndarray): Flat array of hyperparameters, sorted as [length_scale, sigma_f, sigma_n].
        x (ndarray): Input training points.
        y (ndarray): Observed output data.
        xtest (ndarray): Prediction points.
        kernel (function): Function to build the covariance matrix.
        return_full_cov (bool): If True, returns the full covariance matrix, otherwise only its diagonal.
        neig (int): Initial number of eigenvalues to be computed during the inversion of the covariance matrix.

    Returns:
        tuple: A tuple containing:
            - fstar (ndarray): Posterior mean.
            - vstar (ndarray): Posterior covariance matrix or its diagonal.
    """

    if len(hyp) < 3:
        hyp[2] = 0  # sigma_n
    Ky = kernel(x, x, hyp[:-2], *hyp[-2:])
    Kstar = kernel(x, xtest, hyp[:-2], hyp[-2])
    Kstarstar = kernel(xtest, xtest, *hyp)
    Kyinv = invert(Ky, neig, 1e-6*hyp[-1])
    fstar = Kstar.T @ (Kyinv @ y)
    vstar = Kstarstar - Kstar.T @ (invert(Ky) @ Kstar)
    vstar = np.maximum(vstar, 1e-10)  # Assure a positive variance
    if not return_full_cov:
        vstar = np.diag(vstar)
    return fstar, vstar



"""
Testcases for mockup simulations with active learning:
- input variables are ActiveLearning instances
- 1D function
    - simple active learning of one variable
- 2D function (Rosenbrock)
    - active learning of one variable while the other is a Halton sequence
    - simultanious active learning of both variables
"""

from profit.config import Config
from profit.sur import Surrogate
from os import path, remove, chdir, getcwd
from subprocess import run
from numpy import allclose
from shutil import rmtree
from pytest import fixture


@fixture(autouse=True)
def chdir_pytest():
    pytest_root_dir = getcwd()
    chdir(path.dirname(path.abspath(__file__)))
    yield
    chdir(pytest_root_dir)


# Allow a large range of parameters, just ensure that it is approximately at the same scale
PARAM_RTOL = 2
TIMEOUT = 30  # seconds


def clean(config):
    """Delete run directories and input/outpt files after the test."""

    for krun in range(config.get('ntrain')):
        single_dir = path.join(config.get('run_dir'), f'run_{krun:03d}')
        if path.exists(single_dir):
            rmtree(single_dir)
    if path.exists(config['run']['interface'].get('path')):
        remove(config['run']['interface'].get('path'))
    if path.exists(config['files'].get('input')):
        remove(config['files'].get('input'))
    if path.exists(config['files'].get('output')):
        remove(config['files'].get('output'))


def test_1D():
    """Test a simple function f(u) = cos(10*u) + u."""

    config_file = 'study_1D/profit_1D.yaml'
    config = Config.from_file(config_file)
    model_file = './study_1D/model_1D_Custom.hdf5'
    try:
        run(f"profit run {config_file}", shell=True, timeout=TIMEOUT)
        sur = Surrogate.load_model(model_file)
        assert sur.get_label() == 'Custom'
        assert sur.trained
        assert sur.kernel.__name__ == 'RBF'
        assert allclose(sur.hyperparameters['length_scale'], 0.15975022, rtol=PARAM_RTOL)
        assert allclose(sur.hyperparameters['sigma_f'], 0.91133526, rtol=PARAM_RTOL)
        assert allclose(sur.hyperparameters['sigma_n'], 0.00014507, rtol=PARAM_RTOL)
    finally:
        clean(config)
        if path.exists(model_file):
            remove(model_file)


def test_2D():
    """Test a Rosenbrock 2D function with two random inputs."""

    config_file = 'study_2D/profit_2D.yaml'
    config = Config.from_file(config_file)
    model_file = './study_2D/model_2D_Custom.hdf5'
    try:
        run(f"profit run {config_file}", shell=True, timeout=TIMEOUT)
        sur = Surrogate.load_model(model_file)
        assert sur.get_label() == 'Custom'
        assert sur.trained
        assert sur.kernel.__name__ == 'RBF'
        assert sur.ndim == 2
        assert allclose(sur.hyperparameters['length_scale'], 0.96472754, rtol=PARAM_RTOL)
        assert allclose(sur.hyperparameters['sigma_f'], 15.02288291, rtol=PARAM_RTOL)
        assert allclose(sur.hyperparameters['sigma_n'], 8.83125694e-06, rtol=PARAM_RTOL)
    finally:
        clean(config)
        if path.exists(model_file):
            remove(model_file)



"""High-dimensional output

This module concerns the following use-case: we make a parameter study over some
input parameters x and the domain code yields an output vector y contains many
entries. This is typically the case, when y is function-valued, i.e. depends
on an indenpendent variable t, or even "pixel-valued" output is produced,
like on a 2D map (like 1024x1024). Before fitting an input-output relation,
one has to reduce the output to a manageable number of dimensions.
This means extracting the relevant features.

Imagine the following idea: instead of viewing the output as many 1D outputs,
we look at it as a single vector in a high-dimensional space. Linear
combinations between vectors give new vectors. We can choose y from our
training data to span a basis in this vector space. Reducing this basis to an
appropriate set of orthonormal entries is done via the Karhunen-Loeve expansion.

"""

import numpy as np
from numpy.linalg import eigh


class KarhunenLoeve:
    r"""Linear dimension reduction by the Karhunen-Loeve expansion.
    This is efficient if the number of training samples ntrain is
    smaller than the number of support points N in independent variables.

    We want to write the i-th output function $y_i$
    as a linear combination of the other $y_j$ with, e.g. $j=1,2,3$
    $$
    y_i = a_1 y_1 + a_2 y_2 + a_3 y_3
    $$

    This can be done by projection with inner products:

    $$
    \begin{align}
    y_1 \cdot y_i &= a_1 y_1 \cdot y_1+a_2 y_1 \cdot y_2+a_3 y_1 \cdot y_3 \\
    y_2 \cdot y_i &= a_1 y_2 \cdot y_1+a_2 y_2 \cdot y_2+a_3 y_2 \cdot y_3 \\
    y_3 \cdot y_i &= a_1 y_3 \cdot y_1+a_2 y_3 \cdot y_2+a_3 y_3 \cdot y_3
    \end{align}
    $$

    We see that we have to solve a linear system with the
    collocation matrix $M_{ij} = y_i \cdot y_j$. To find the most relevant
    features and reduce dimensionality, we use only the highest eigenvalues.
    We center the data around the mean, i.e. subtract `ymean` before.

    Parameters:
        ytrain: ntrain sample vectors of length N.
        tol: Absolute cutoff tolerance of eigenvalues.
    """
    def __init__(self, ytrain, tol=1e-2):
        self.tol = tol
        self.ymean = np.mean(ytrain, 0)
        self.dy = ytrain - self.ymean
        w, Q = eigh(self.dy @ self.dy.T)
        condi = w>tol
        self.w = w[condi]
        self.Q = Q[:, condi]

    def project(self, y):
        """
        Parameters:
            y: ntest sample vectors of length N.

        Returns:
            Expansion coefficients of y in eigenbasis.
        """
        ntrain = self.dy.shape[0]
        ntest = y.shape[0]
        b = np.empty((ntrain, ntest))
        for i in range(ntrain):
            b[i,:] = (y - self.ymean) @ self.dy[i]
        return np.diag(1.0/self.w) @ self.Q.T @ b

    def lift(self, z):
        """
        Parameters:
            z: Expansion coefficients of y in eigenbasis.

        Returns:
            Reconstructed ntest sample vectors of length N.
        """
        return self.ymean + (self.dy.T @ (self.Q @ z)).T

    def features(self):
        """
        Returns:
            neig feature vectors of length N.
        """
        return self.dy.T @ self.Q


class PCA:
    """Linear dimension reduction by principle component analysis (PCA).
    This is efficient if the number of training samples ntrain is
    larger than the number of support points in independent variables.

    Parameters:
        ytrain: ntrain sample vectors of length N.
        tol: Absolute cutoff tolerance of eigenvalues.
    """
    def __init__(self, ytrain, tol=1e-2):
        self.tol = tol
        self.ymean = np.mean(ytrain, 0)
        self.dy = ytrain - self.ymean
        w, Q = eigh(self.dy.T @ self.dy)
        condi = w>tol
        self.w = w[condi]
        self.Q = Q[:, condi]

    def project(self, y):
        """
        Parameters:
            y: ntest sample vectors of length N.

        Returns:
            Expansion coefficients of y in eigenbasis.
        """
        return (y - self.ymean) @ self.Q

    def lift(self, z):
        """
        Parameters:
            z: Expansion coefficients of y in eigenbasis.

        Returns:
            Reconstructed ntest sample vectors of length N.
        """
        return self.ymean + (self.Q @ z.T).T

    def features(self):
        """
        Returns:
            neig feature vectors of length N.
        """
        return self.Q



import dash
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output
from textwrap import dedent as d

import pandas as pd

#df = pd.read_hdf(r'..\examples\algae\MC\mc_out_allyears_sigma1.h5')
df = pd.read_csv(r'..\examples\algae\MC\mc_out_allyears_sigma1.dat', delim_whitespace=True)
data = df.values
param = data[:,0:8]


def generate_table():
    data = dict()
    for l in range(4):
        for k in range(2):
            data[k+2*l] = [
                {
                    'x': df.iloc[::100, k+2*l],
                    'type': 'histogram'
                }
            ]
    return html.Table(
        [html.Tr([
            html.Td(dcc.Graph(id='hist{}'.format(k+2*l),
            figure = {'data': data[k+2*l], 'layout': {}}),
            style={'width': '500px'}) for k in range(2)
        ], style={'height': '300px'}) for l in range(4)],

    )


external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']

app = dash.Dash(__name__, external_stylesheets=external_stylesheets)

app.layout = html.Div(children=[
    html.H1(children='Conditional probability distributions'),

    html.Div([
        dcc.Markdown(d("""
            **Click** on bars in one histogram to plot conditional
            probability distribution for other variables.
        """)),
        html.Button('Reset', id='reset', n_clicks_timestamp=0),
        #html.Pre(id='click-data'),
    ]),

    generate_table()
])

#@app.callback(
#    Output('click-data', 'children'),
#    [Input('hist0', 'clickData')])
#def display_click_data(clickData):
    #return json.dumps(clickData['points'][0], indent=2)
#    try:
#        return json.dumps(clickData['points'][0]['binNumber'])
#    except:
#        return 0

def gen_callback(k):
    from numpy import array
    def update_figure(*args):
        defaultdata = {
            'data': [
                    {
                        'x': df.iloc[::100, k],
                        'type': 'histogram'
                    }
                ]
            }

        # Find out who triggered the callback,
        # see https://github.com/plotly/dash/issues/291
        ctx = dash.callback_context

        # Update was not triggered at all
        if not ctx.triggered:
            return defaultdata
        trigger = ctx.triggered[0]

        # Reset button triggered update
        if trigger['prop_id'] == 'reset.n_clicks':
            return defaultdata

        # Other component triggered update
        try:
            reducedset = array(trigger['value']['points'][0]['pointNumbers'])
        except:
            return defaultdata
        return {
            'data': [
                    {
                        'x': df.iloc[reducedset, k],
                        'type': 'histogram'
                    }
                ]
            }
    return update_figure


for k in range(8):
    app.callback(
        output = Output('hist{}'.format(k), 'figure'),
        inputs = (
            [Input('hist{}'.format(l), 'clickData') for l in range(8)]
            + [Input('reset', 'n_clicks')]))(gen_callback(k))


if __name__ == '__main__':
    app.run_server(debug=True, host='0.0.0.0')



from os import path, chdir, listdir
from .util import save, load


def read_input(filename):
    """ Loads data from input file into a numpy array. """
    data = load(filename)
    return data.view((float, len(data.dtype.names))).T


def collect_output(config, default_interface=False):
    """ Collects simulation results from each run directory into a single output file. """

    from numpy import zeros, nan
    from importlib.util import spec_from_file_location, module_from_spec
    from tqdm import tqdm

    if not default_interface:
        try:
            spec = spec_from_file_location('interface', config['interface'])
            interface = module_from_spec(spec)
            spec.loader.exec_module(interface)
        except FileNotFoundError:
            raise ImportError("Could not load interface {}".format(config['interface']))
    else:
        try:
            name = [f for f in listdir(path.join(config['run_dir'], '000')) if f.endswith('out')][0]
        except IndexError:
            name = None
        interface = DefaultInterface(name)

    # TODO: do this in less code?
    # Header for output
    header = []
    for out, values in config['output'].items():
        if not values['range']:
            header.append("{f}".format(f=out))
        else:
            for dependent, rng in values['range'].items():
                for number in rng.flatten():
                    header.append("{f}({x}={n})".format(f=out, x=dependent, n=round(number, 2)))

    # Get vector output
    nout = 0
    for v in config['output'].values():
        for rng in v['range'].values():
            nout += rng.size

    dtypes = [(key, float) for key in config['output'].keys()]
    data = zeros((config['ntrain'], max(int(nout), 1)), dtype=dtypes)

    kruns = tqdm(range(config['ntrain']))
    for krun in kruns:
        # .zfill(3) is an option that forces krun to have 3 digits
        run_dir_single = path.join(config['run_dir'], str(krun).zfill(3))
        print(run_dir_single)
        try:
            chdir(run_dir_single)
            # TODO: make get_output run with parameters e.g. config['interface']['params'] as *args
            # Interface should return a tuple or list if more than one output variable.
            alldata = interface.get_output()
            for i, key in enumerate(data.dtype.names):
                data[key][krun, :] = alldata[i] if isinstance(alldata, (tuple, list)) else alldata
        except:
            data[krun, :] = nan
        finally:
            chdir(config['run_dir'])

    save(config['files']['output'], data, ' '.join(header))



"""
Created: Tue May 28 15:40:29 2019
@author: Christopher Albert <albert@alumni.tugraz.at>
"""

import yaml


with open('../examples/mockup/suruq.yaml', 'r') as document:
    config = yaml.load(document)

print(yaml.dump(config))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug 19 03:47:54 2020

@author: manal khallaayoune
"""

import numpy as np
import GPy
from GPy.kern.src.stationary import Stationary


class ExpSin(Stationary):
    """
    Exponential of Sinus kernel: 
        Product of 1D Exponential of Sinus kernels

    .. math::

        &k(x,x')_i = \sigma^2 \prod_{j=1}^{dimension} \exp \\bigg( - \\frac{ \sin ( x_{i,j}-x_{i,j}' ) ^2}{2 \ell_j^2} \\bigg)
        
        &x,x' \in \mathcal{M}_{n,dimension}
        
        &k \in \mathcal{M}_{n,n}

    """
    def __init__(self, input_dim, variance=1., lengthscale=None, ARD=False, active_dims=None, name='ExpSin'):
        super(ExpSin, self).__init__(input_dim, variance, lengthscale, ARD, active_dims, name)

    def K_of_r(self, dist):
        n = dist.shape[2]
        s = 0
        for k in range(n):
            s+= np.sin(dist[:,:,k])**2
        return self.variance * np.exp(-s/(2*self.lengthscale**2))
    
    def K(self, X, X2):
        dist = X[:,None,:]-X2[None,:,:]
        return self.K_of_r(dist)

    def dK_dr(self,dist,dimX):
        K = self.K_of_r(dist)
        d1 = dist[:,:,dimX]
        return -K*np.sin(d1)*np.cos(d1)/self.lengthscale**2
    
    def dK_dX(self, X, X2, dimX):
        dist = X[:,None,:]-X2[None,:,:]
        dK_dr = self.dK_dr(dist,dimX)
        return dK_dr
    
    def dK_dX2(self,X,X2,dimX2):
        return -self.dK_dX(X,X2, dimX2)
    
    def dK2_dXdX2(self, X, X2, dimX, dimX2):
        dist = X[:,None,:]-X2[None,:,:]
        K = self.K_of_r(dist)
        dK_dX = self.dK_dX(X,X2,dimX)
        lengthscale2inv = np.ones((X.shape[1]))/(self.lengthscale**2)
        l1 = lengthscale2inv[dimX]
        l2 = lengthscale2inv[dimX2]
        d1 = dist[:,:,dimX]
        d2 = dist[:,:,dimX2]
        s1 = np.sin(d1)
        c1 = np.cos(d1)
        return (dimX!=dimX2)*dK_dX*l2*np.sin(d2)*np.cos(d2) + (dimX==dimX2)*l1*(dK_dX*s1*c1+K*c1**2-K*s1**2)
    

# #%%
# import pytest
# import func
# import matplotlib.pyplot as plt

# def test_kern_Cosine():
#     x0train = np.linspace(-5,5,100).reshape(-1,1)
#     x1train = np.linspace(-2,2,100).reshape(-1,1)
#     x2train = np.linspace(0,9,100).reshape(-1,1)
#     x3train = np.linspace(8,18,100).reshape(-1,1)
#     xtrain = np.hstack((x0train, x1train))#, x2train))#, x3train))
    
#     # k0 = ExpSin(2)#, active_dims=1)
#     # k0_K = k0.K(x0train,x1train)
#     # dk0 = k0.dK_dX(x0train, x1train, 0)
#     # dk0_2 = k0.dK2_dXdX2(x0train, x1train, 0, 0)
    
#     k0 = ExpSin(2)#, active_dims=1)
#     k0_K = k0.K(xtrain,xtrain)
#     dk0 = k0.dK_dX(xtrain, xtrain, 1)
#     dk0_2 = k0.dK2_dXdX2(xtrain, xtrain, 1, 1)
    
#     l = np.array([1, 1])
    
#     K = np.zeros((len(x0train),len(x0train)))
#     for i in range(K.shape[0]):
#         for j in range(K.shape[1]):
#             K[i,j] = func.f_kern(xtrain[i,0], xtrain[i,1], xtrain[j,0], xtrain[j,1], l)
    
#     dK = np.zeros((len(x0train),len(x0train)))
#     for i in range(dK.shape[0]):
#         for j in range(dK.shape[1]):
#             dK[i,j] = func.dkdy(xtrain[i,0], xtrain[i,1], xtrain[j,0], xtrain[j,1], l)
    
#     dK2 = np.zeros((len(x0train),len(x0train)))
#     for i in range(dK2.shape[0]):
#         for j in range(dK2.shape[1]):
#             dK2[i,j] = func.d2kdydy0(xtrain[i,0], xtrain[i,1], xtrain[j,0], xtrain[j,1], l)

#     plt.figure()
#     plt.imshow(K)
#     plt.title('K sympy' '\n' 'order 2')
#     plt.figure()
#     plt.imshow(k0_K)
#     plt.title('K' '\n' 'order 2')
#     plt.figure()
#     plt.plot(K[49,:])
#     plt.title('K sympy' '\n' 'order 2')
#     plt.figure()
#     plt.plot(k0_K[49,:])
#     plt.title('K' '\n' 'order 2')
    
#     plt.figure()
#     plt.imshow(dK)
#     plt.title('1st derivative sympy' '\n' 'order 2')
#     plt.figure()
#     plt.imshow(dk0)
#     plt.title('1st derivative' '\n' 'order 2')
#     plt.figure()
#     plt.plot(dK[49,:])
#     plt.title('1st derivative sympy' '\n' 'order 2')
#     plt.figure()
#     plt.plot(dk0[49,:])
#     plt.title('1st derivative' '\n' 'order 2')
    
#     plt.figure()
#     plt.imshow(dK2)
#     plt.title('2nd derivative sympy' '\n' 'order 2')
#     plt.figure()
#     plt.imshow(dk0_2)
#     plt.title('2nd derivative' '\n' 'order 2')
#     plt.figure()
#     plt.plot(dK2[49,:])
#     plt.title('2nd derivative sympy' '\n' 'order 2')
#     plt.figure()
#     plt.plot(dk0_2[49,:])
#     plt.title('2nd derivative' '\n' 'order 2')

#     print(np.isclose(k0_K, K, rtol=1e-6),'\n')
#     print(np.isclose(dk0, dK, rtol=1e-6),'\n')
#     print(np.isclose(dk0_2, dK2, rtol=1e-6))
#     # print(dk0_2,'\n')
#     # print(dk0)
#     # return dist
    

# test_kern_Cosine()











#%%
import numpy as np
import GPy

k0 = GPy.kern.RBF(1, active_dims=0)
k1 = GPy.kern.RBF(1, active_dims=1)

k0_der = GPy.kern.DiffKern(k0, 0)

# Extended class for product kernel,
# can be merged with Prod class when finished

class ProdExtended(GPy.kern.Prod):
    # def K(self,X,X2):
    #     parts = self.parts[:]
    #     All = parts[:]
    #     n = len(All)
    #     m = len(X2)#X.shape[1]
    #     # if (n==m):
    #     K = np.ones((m,m))
    #     for i in range(m):
    #         K[:,:] *= All[i].K(X[:,i],X2[:,i])
    #     return np.prod(K,0)
    #     # else:
    #     #     raise(NotImplementedError)
    
    # def dK_dr(self,r):
    #     raise(NotImplementedError)
    
    def dK_dX(self, X, X2, dimX):
        # Apply product rule to kprod = k0*k1
        # so    d(kprod/dx0) = dk0/dx0*k1
        # and   d(kprod/dx1) = k0*dk1/dx1
        other = self.parts[:] 
        All = other[:]
        n = len(All)
        nX = X.shape[0]
        nX2 = X2.shape[0]
        diffpart = other.pop(dimX)
        K_all = np.zeros((n,nX,nX2))
        for k in range(n):
            if (k!=dimX):
                K_all[k,:,:] = All[k].K(X,X2)
            else:
                K_all[k,:,:] = diffpart.dK_dX(X,X2,dimX)
        # return k0.dK_dX(X,X2,dimX)*(k1.K(X,X2))
        return np.prod(K_all,0)
        
    def dK_dX2(self,X,X2,dimX2):
        # return k0.K(X,X2)*k1.dK_dX(X2,X,dimX2)
        # return k0.K(X,X2)*k1.dK_dX2(X,X2,dimX2)
        return -self.dK_dX(X,X2,dimX2)

    def dK2_dXdX2(self, X, X2, dimX, dimX2):
        # Apply product rule to kprod = k0*k1
        # so    d²(kprod/dx0²) = d²k0/dx0²*k1
        # and   d²(kprod/dx1²) = k0*d²k1/dx1²
        # so    d²(kprod/dx0dx1) = dk0/dx0*dk1/dx1
        # and   d²(kprod/dx1dx0) = dk0/dx0*dk1/dx1
        # raise(NotImplementedError)
        # if (dimX==dimX2 & dimX==k0.active_dims):
        #     return k0.dK2_dXdX2(X,X2,dimX,dimX)*k1.K(X,X2)
        # elif (dimX==dimX2 & dimX==k1.active_dims):
        #     return k0.K(X,X2)*k1.dK2_dXdX2(X,X2,dimX,dimX)
        # return self.dK_dX(X,X2,dimX)*self.dK_dX2(X,X2,dimX2)
        
        # other = kprod.parts[:] 
        # if (dimX==dimX2):
        #     diffpart = other.pop(dimX)
        #     dK2dXdX2 = diffpart.dK2_dXdX2(X,X2,dimX,dimX)
        #     All = other
        #     All[dimX] = dK2dXdX2
        #     return np.prod(All)
        # diffpart1 = other.pop(dimX)
        # diffpart2 = other.pop(dimX2)
        # dK2dXdX2_1 = diffpart.dK_dX(X,X2,dimX)
        # dK2dXdX2_2 = diffpart.dK_dX(X,X2,dimX2)
        # All = other
        # All[dimX] = dK2dXdX2_1
        # All[dimX] = dK2dXdX2_1
        # return np.prod(All)
        
        other = self.parts[:] 
        All = other[:]
        n = len(All)
        nX = X.shape[0]
        nX2 = X2.shape[0]
        if (dimX==dimX2):
            diffpart = other.pop(dimX)
            K_all = np.zeros((n,nX,nX2))
            for k in range(n):
                if (k!=dimX):
                    K_all[k,:,:] = All[k].K(X,X2)
                else:
                    K_all[k,:,:] = diffpart.dK2_dXdX2(X,X2,dimX,dimX)
        else:
            diffpart1 = other.pop(dimX)
            diffpart2 = All[dimX2]
            K_all = np.zeros((n,nX,nX2))
            for k in range(n):
                if (k==dimX):
                    K_all[k,:,:] = diffpart1.dK_dX(X,X2,dimX)
                elif (k==dimX2):
                    K_all[k,:,:] = diffpart2.dK_dX(X,X2,dimX2)
                else:
                    K_all[k,:,:] = All[k].K(X,X2)
        return ((dimX==dimX2)+(dimX!=dimX2)*(-1.))*np.prod(K_all,0)

# %%
# kprod = ProdExtended((k0, k1))

# x0train = np.array([0.0, 1.0, 0.0]).reshape(-1,1)
# x1train = np.array([0.0, 0.0, 1.0]).reshape(-1,1)
# xtrain = np.hstack((x0train, x1train))


# print('Training points:')
# print(xtrain)
# print()

# print('K0 = ')
# print(k0.K(x0train, x0train))
# print(k0.dK_dX(x0train, x0train, 0))
# print(k0.dK2_dXdX2(x0train, x0train, 0, 0))
# print()

# print('K1 = ')
# print(k1.K(xtrain, xtrain))  # Need 2D vectors, here, as active_dims=1
# print()

# print('Prod K = ')
# print(kprod.K(xtrain, xtrain))
# print()

# #%% TODO: test 1st and 2nd derivatives based on parts of Prod kernel
# X = xtrain
# X2 = xtrain
# dimX = 0
# # Apply product rule to kprod = k0*k1
# # so    d(kprod/dx0) = dk0/dx0*k1
# # and   d(kprod/dx1) = k0*dk1/dx1

# other = kprod.parts[:]      # to store all parts except the dimX one
# diffpart = other.pop(dimX)  # removes dimX and returns it as diffpart

# print('Part to differentiate:')
# print(diffpart)  # should give k0 here
# print('Other parts')
# print(other[0])  # should give k1 here

# dK_dX_diffpart = diffpart.dK_dX(X, X2, dimX)
# K_other = [k.K(X, X2) for k in other]
# result = dK_dX_diffpart*np.prod(K_other)

# print('Derivative factor:')
# print(dK_dX_diffpart)
# print('Other factors:')
# print(K_other)
# print('Overall result:')
# print(result)

# #%% This is not implemented yet, TODO in class
# print('Derivatives Prod K')
# print('kprod.dK_dX',kprod.dK_dX(xtrain, xtrain, 0),'\n')
# print('kprod.dK_dX2',kprod.dK_dX2(xtrain, xtrain, 1),'\n')
# print('kprod.dK2_dXdX2',kprod.dK2_dXdX2(xtrain, xtrain, 0, 0),'\n')
# # print(k0.dK2_dXdX2(xtrain,xtrain,0,0)*k1.K(xtrain,xtrain))
# print(kprod.dK2_dXdX2(xtrain, xtrain, 1, 1))
# # print(kprod.dK2_dXdX2(xtrain, xtrain, 1, 0))
# # print(kprod.dK2_dXdX2(xtrain, xtrain, 1, 1))

# # print(k0.K(xtrain,xtrain)*k1.dK_dX(xtrain,xtrain,1))


# #%%
# import pytest
# import func
# import matplotlib.pyplot as plt
# # from cosine_prod import Cosine_prod
# from expsin_gpy import ExpSin


# def test_kern_ProdExtended():
#     x0train = np.linspace(-5,5,100).reshape(-1,1)
#     x1train = np.linspace(-2,2,100).reshape(-1,1)
#     x2train = np.linspace(0,9,100).reshape(-1,1)
#     x3train = np.linspace(8,18,100).reshape(-1,1)
#     xtrain = np.hstack((x0train, x0train))#, x2train))#, x1train))
#     xtrain1 = np.hstack((x1train, x0train))#, x2train))#, x3train))
#     xtrain2 = np.hstack((x0train, x0train))#, x2train))#, x3train))
#     xtrain3 = np.hstack((x1train, x1train))#, x2train))#, x3train))
    
#     # k1 = Cosine_prod(1,active_dims=0)
#     # k2 = Cosine_prod(1,active_dims=1)
#     # k4 = Cosine_prod(1,active_dims=2)
#     k1 = ExpSin(1)#,active_dims=0)
#     k2 = GPy.kern.RBF(1,active_dims=1)
#     # k4 = ExpSin(1,active_dims=2)
#     # k1 = GPy.kern.RBF(1)
#     # k2 = GPy.kern.RBF(1)
#     # k2 = ExpSin(1)
#     k0 = ProdExtended((k1,k2))
#     # k0_K = k0.K(xtrain,xtrain1)
#     # k0 = ProdExtended((k1,k2))
#     k0_K = k0.K(xtrain,xtrain)

#     dk0 = k0.dK_dX(xtrain, xtrain, 0)
#     dk0_2 = k0.dK2_dXdX2(xtrain, xtrain, 0, 0)
    
#     # # k0 = Sinus_prod(2)#, active_dims=1)
#     # # k0_K = k0.K(xtrain,xtrain)
#     # # dk0 = k0.dK_dX(xtrain, xtrain, 1)
#     # # dk0_2 = k0.dK2_dXdX2(xtrain, xtrain, 1, 0)
    
#     l = np.array([1, 1])
    
#     K = np.zeros((len(x0train),len(x0train)))
#     for i in range(K.shape[0]):
#         for j in range(K.shape[1]):
#             K[i,j] = func.f_kern(xtrain[i,0], xtrain[i,1], xtrain[j,0], xtrain[j,1], l)
    
#     dK = np.zeros((len(x0train),len(x0train)))
#     for i in range(dK.shape[0]):
#         for j in range(dK.shape[1]):
#             dK[i,j] = func.dkdx(xtrain[i,0], xtrain[i,1], xtrain[j,0], xtrain[j,1], l)
    
#     dK2 = np.zeros((len(x0train),len(x0train)))
#     for i in range(dK2.shape[0]):
#         for j in range(dK2.shape[1]):
#             dK2[i,j] = func.d2kdxdx0(xtrain[i,0], xtrain[i,1], xtrain[j,0], xtrain[j,1], l)
    
#     # # l = np.array([1, 1])
    
#     # k3 = ExpSin(2)
#     # # k3 = GPy.kern.RBF(2)
#     # # K = k3.K(xtrain2,xtrain3)
#     # # k3 = GPy.kern.RBF(2)
#     # K = k3.K(xtrain,xtrain)

#     # dK = k3.dK_dX(xtrain,xtrain,1)
#     # dK2 = k3.dK2_dXdX2(xtrain,xtrain,1, 0)
    
#     # k1 = GPy.kern.RBF(1, active_dims=0)  # SqExp in first dimension
#     # k2 = GPy.kern.RBF(1, active_dims=1)  # SqExp in second dimension
#     # k0 = ProdExtended((k1, k2))
#     # k0_K = k0.K(xtrain,xtrain)
#     # dk0 = k0.dK_dX(xtrain, xtrain, 1)
#     # dk0_2 = k0.dK2_dXdX2(xtrain, xtrain, 0, 1)
    
#     # k3 = GPy.kern.RBF(2)  # SqExp in 2D for comparison
#     # K = k3.K(xtrain,xtrain)
#     # dK = k3.dK_dX(xtrain,xtrain,1)
#     # dK2 = k3.dK2_dXdX2(xtrain,xtrain,1,0)
    
#     # x0train = np.array([0.0, 1.0, 0.0]).reshape(-1,1)
#     # x1train = np.array([0.0, 0.0, 1.0]).reshape(-1,1)
#     # xtrain = np.hstack((x0train, x1train))

#     # print('Prod K = ')
#     # print(kprod.K(xtrain, xtrain))
#     # print()
    
#     # print('Reference K = ')
#     # print(k01.K(xtrain, xtrain))
#     # print()
    
#     # k0= Cosine_prod(1)
#     # k0_K = k0.K(xtrain,xtrain)
#     # print(k0_K.shape)
    
#     # k1 = Cosine_prod(1)
#     # K = k1.K(x0train,x0train)
#     # print(K.shape)
    
    
#     # K = np.zeros((len(x0train),len(x0train)))
#     # for i in range(K.shape[0]):
#     #     for j in range(K.shape[1]):
#     #         K[i,j] = func.f_kern(xtrain[i,0], xtrain[i,1], xtrain[j,1], xtrain[j,1], l)
    
#     # dK = np.zeros((len(x0train),len(x0train)))
#     # for i in range(dK.shape[0]):
#     #     for j in range(dK.shape[1]):
#     #         dK[i,j] = func.dkdx(xtrain[i,0], xtrain[i,1], xtrain[j,1], xtrain[j,1], l)
    
#     # dK2 = np.zeros((len(x0train),len(x0train)))
#     # for i in range(dK2.shape[0]):
#     #     for j in range(dK2.shape[1]):
#     #         dK2[i,j] = func.d2kdxdx0(xtrain[i,0], xtrain[i,1], xtrain[j,1], xtrain[j,1], l)

#     plt.figure()
#     plt.imshow(K)
#     plt.title('K sympy' '\n' 'order 2')
#     plt.figure()
#     plt.imshow(k0_K)
#     plt.title('K' '\n' 'order 2')
#     plt.figure()
#     plt.plot(K[49,:])
#     plt.title('K sympy' '\n' 'order 2')
#     plt.figure()
#     plt.plot(k0_K[49,:])
#     plt.title('K' '\n' 'order 2')
    
#     plt.figure()
#     plt.imshow(dK)
#     plt.title('1st derivative sympy' '\n' 'order 2')
#     plt.figure()
#     plt.imshow(dk0)
#     plt.title('1st derivative' '\n' 'order 2')
#     plt.figure()
#     plt.plot(dK[49,:])
#     plt.title('1st derivative sympy' '\n' 'order 2')
#     plt.figure()
#     plt.plot(dk0[49,:])
#     plt.title('1st derivative' '\n' 'order 2')
    
#     plt.figure()
#     plt.imshow(dK2)
#     plt.title('2nd derivative sympy' '\n' 'order 2')
#     plt.figure()
#     plt.imshow(dk0_2)
#     plt.title('2nd derivative' '\n' 'order 2')
#     plt.figure()
#     plt.plot(dK2[49,:])
#     plt.title('2nd derivative sympy' '\n' 'order 2')
#     plt.figure()
#     plt.plot(dk0_2[49,:])
#     plt.title('2nd derivative' '\n' 'order 2')

#     print(np.isclose(k0_K, K, rtol=1e-6),'\n')
#     print(np.isclose(dk0, dK, rtol=1e-6),'\n')
#     print(np.isclose(dk0_2, dK2, rtol=1e-6))
    

# test_kern_ProdExtended()



"""proFit default runner

in development (Mar 2021)
Goal: a class to manage and deploy runs
"""

import os
import shutil
import logging
from abc import ABC, abstractmethod  # Abstract Base Class
from collections.abc import MutableMapping

from .worker import Preprocessor, Postprocessor, Worker
from profit.util import load_includes, params2map, spread_struct_horizontal, flatten_struct

import numpy as np


class RunnerInterface:
    interfaces = {}  # ToDo: rename to registry?
    internal_vars = [('DONE', np.bool8), ('TIME', np.uint32)]

    def __init__(self, config, size, input_config, output_config, *, logger_parent: logging.Logger = None):
        self.config = config  # base_config['run']['interface']
        self.logger = logging.getLogger('Runner Interface')
        if logger_parent is not None:
            self.logger.parent = logger_parent

        self.input_vars = [(variable, spec['dtype']) for variable, spec in input_config.items()]
        self.output_vars = [(variable, spec['dtype'], spec['shape']) for variable, spec in output_config.items()]

        self.input = np.zeros(size, dtype=self.input_vars)
        self.output = np.zeros(size, dtype=self.output_vars)
        self.internal = np.zeros(size, dtype=self.internal_vars)

    def poll(self):
        self.logger.debug('polling')

    def clean(self):
        self.logger.debug('cleaning')

    @classmethod
    def handle_config(cls, config, base_config):
        pass

    @classmethod
    def register(cls, label):
        def decorator(interface):
            if label in cls.interfaces:
                raise KeyError(f'registering duplicate label {label} for Interface')
            cls.interfaces[label] = interface
            return interface
        return decorator

    def __class_getitem__(cls, item):
        return cls.interfaces[item]


# === Runner === #


class Runner(ABC):
    systems = {}

    # for now, implement the runner straightforward with less overhead
    # restructuring is always possible
    def __init__(self, interface_class, run_config, base_config):
        self.base_config = base_config
        self.run_config = run_config
        self.config = self.run_config['runner']
        self.logger = logging.getLogger('Runner')
        self.interface: RunnerInterface = interface_class(self.run_config['interface'], self.base_config['ntrain'],
                                                          self.base_config['input'], self.base_config['output'],
                                                          logger_parent=self.logger)

        self.runs = {}  # run_id: (whatever data the system tracks)
        self.next_run_id = 0
        self.env = os.environ.copy()
        self.env['PROFIT_BASE_DIR'] = self.base_config['base_dir']
        self.env['PROFIT_CONFIG_PATH'] = base_config['config_path']  # ToDo better way to pass this?

    @classmethod
    def from_config(cls, config, base_config):
        child = cls[config['runner']['class']]
        interface_class = RunnerInterface[config['interface']['class']]
        return child(interface_class, config, base_config)

    def fill(self, params_array, offset=0):
        for r, row in enumerate(params_array):
            mapping = params2map(row)
            for key, value in mapping.items():
                self.interface.input[key][r + offset] = value

    @abstractmethod
    def spawn_run(self, params=None, wait=False):
        """spawn a single run

        :param params: a mapping which defines input parameters to be set
        :param wait: whether to wait for the run to complete
        """
        mapping = params2map(params)
        for key, value in mapping.items():
            self.interface.input[key][self.next_run_id] = value

    def spawn_array(self, params_array, blocking=True):
        if not blocking:
            raise NotImplementedError
        for params in params_array:
            self.spawn_run(params, wait=True)

    @abstractmethod
    def check_runs(self):
        pass

    @abstractmethod
    def cancel_all(self):
        pass

    def clean(self):
        self.interface.clean()

    @property
    def input_data(self):
        return self.interface.input[self.interface.internal['DONE']]

    @property
    def flat_input_data(self):
        return flatten_struct(self.input_data)

    @property
    def output_data(self):
        return self.interface.output[self.interface.internal['DONE']]

    @property
    def structured_output_data(self):
        return spread_struct_horizontal(self.output_data, self.base_config['output'])

    @property
    def flat_output_data(self):
        return flatten_struct(self.output_data)

    @classmethod
    def handle_run_config(cls, base_config, run_config=None):
        """ handle the config dict, fill in defaults & delegate to the children

        Arguments:
            run_config: dict read from config file, ~base_config['run'] usually
            base_config: base dict read from config file

        handles: include, runner, interface, custom, worker
        all other config options need to be handled by the called worker (or other components)

        ToDo: check for correct types & valid paths
        ToDo: give warnings
        """
        run_config = run_config or base_config['run']
        if 'include' not in run_config:
            run_config['include'] = []
        elif isinstance(run_config['include'], str):
            run_config['include'] = [run_config['include']]
        for p, path in enumerate(run_config['include']):
            if not os.path.isabs(path):
                run_config['include'][p] = os.path.abspath(os.path.join(base_config['base_dir'], path))
        load_includes(run_config['include'])

        defaults = {'runner': 'local', 'interface': 'memmap', 'custom': False, 'worker': None}
        for key, default in defaults.items():
            if key not in run_config:
                run_config[key] = default

        if not isinstance(run_config['runner'], MutableMapping):
            run_config['runner'] = {'class': run_config['runner']}
        Runner[run_config['runner']['class']].handle_config(run_config['runner'], base_config)
        if not isinstance(run_config['interface'], MutableMapping):
            run_config['interface'] = {'class': run_config['interface']}
        RunnerInterface[run_config['interface']['class']].handle_config(run_config['interface'], base_config)
        Worker[run_config['worker']].handle_config(run_config, base_config)

    @classmethod
    def handle_config(cls, config, base_config):
        pass

    @classmethod
    def register(cls, label):
        def decorator(interface):
            if label in cls.systems:
                raise KeyError(f'registering duplicate label {label} for Runner')
            cls.systems[label] = interface
            return interface

        return decorator

    def __class_getitem__(cls, item):
        return cls.systems[item]



#%%
import numpy as np
import matplotlib.pyplot as plt
from scipy.sparse.linalg.eigen import eigsh
from sklearn.model_selection import train_test_split

#%%
indata = np.loadtxt(
    '/mnt/c/Users/chral/Downloads/input_sparse.txt',
    skiprows=1)
outdata = np.loadtxt(
    '/mnt/c/Users/chral/Downloads/1998_V7_only_Chl_output_UQ_1998_sparse_proc_000.txt',
    skiprows=1)[:, 1:]
#%%
ndat = outdata.shape[0]    # Number of data
nx = outdata.shape[1] - 1  # Number of independent support points
# (= dimensionality of original output space)

plt.figure()
ktrain = []
for k in range(4):
    ktrain.append(np.random.randint(ndat))
    plt.plot(outdata[ktrain[-1], :])

legend = [f'$u_1={indata[d][1]:1.1f}$' for d in ktrain]

plt.legend(legend)
plt.xlabel('Independent variable t')
plt.ylabel('Output f(t;u)')
plt.title(f'Output f(t;u) for different u')
#%%
utrain, utest, ytrain, ytest = train_test_split(
    indata, outdata, test_size=0.95)

ntrain = utrain.shape[0]
ntest = utest.shape[1]

ymean = np.mean(ytrain, 0)
dytrain = ytrain - ymean
M = dytrain @ dytrain.T

w, Q = eigsh(M, 22)
plt.figure()
plt.semilogy(w[::-1]/w[-1])
plt.xlabel('Number')
plt.ylabel('Relative size')
plt.title('Eigenvalue decay')

neig = 5
Q = Q[:,:-neig-1:-1]
w = w[:-neig-1:-1]

plt.figure()
for i in range(len(w)):
    efunc = Q[:,i] @ dytrain
    plt.plot(-efunc*np.sign(efunc[0]))  # Sign-flipped eigenfunctions
    #plt.plot(efunc)  # Eigenfunctions
plt.title(f'{neig} most significant eigenvectors')
plt.xlabel('Independent variable t')
plt.ylabel('Eigenvectors')
#%% Testing
ktest = np.random.randint(ntest)

b = np.empty(ntrain)
for i in range(ntrain):
    b[i] = (ytest[ktest,:] - ymean) @ dytrain[i]

a = Q @ (np.diag(1.0/w) @ (Q.T @ b))  # Brackets to avoid production of matrices

plt.figure()
plt.plot(ymean + a @ dytrain, '--')
plt.plot(ytest[ktest,:])
plt.xlabel('Independent variable t')
plt.ylabel(f'Output $f(t;u_1={utest[ktest, 1]:1.2f})$')
plt.legend(['Reconstruction', 'Reference'])

# %%

from sklearn.decomposition import PCA, KernelPCA

pca = PCA(n_components=neig)

atrain = pca.fit_transform(ytrain)
atest = pca.transform(ytest)
yfit = pca.inverse_transform(atest)

plt.figure()
plt.fill_between(np.arange(yfit.shape[1]),
    yfit[ktest, :] - 1.96*np.sqrt(pca.noise_variance_),
    yfit[ktest, :] + 1.96*np.sqrt(pca.noise_variance_), alpha=0.2)
plt.plot(yfit[ktest, :], '--')
plt.plot(ytest[ktest,:])
plt.xlabel('Independent variable t')
plt.ylabel(f'Output $f(t;u_1={utest[ktest, 1]:1.2f})$')
plt.legend(['Reconstruction', 'Reference'])


# %%
# kpca = KernelPCA(n_components=neig, kernel='rbf', fit_inverse_transform=True, gamma=10)
# atraink = kpca.fit_transform(ytrain)
# atestk = kpca.transform(ytest)
# yfitk = kpca.inverse_transform(atestk)

# plt.figure()
# plt.plot(yfitk[ktest, :], '--')
# plt.plot(ytest[ktest,:])
# plt.xlabel('Independent variable t')
# plt.ylabel(f'Output $f(t;u_1={utest[ktest, 1]:1.2f})$')
# plt.legend(['Reconstruction', 'Reference'])

# %%



# setup.py
# install script using setuptools / numpy.distutils

import sys
import os
import site

from numpy.distutils.core import Extension, setup

ext_kwargs = {
    'libraries': ['gomp'],
    'extra_f90_compile_args': ['-Wall', '-march=native', '-O2', '-fopenmp',
                               '-g', '-fbacktrace']}

ext_gpfunc = Extension(
    name='profit.sur.backend.gpfunc',
    sources=['profit/sur/backend/gpfunc.f90'],
    **ext_kwargs)

ext_kernels = Extension(
    name='profit.sur.backend.kernels',
    sources=['profit/sur/backend/kernels.f90', 'profit/sur/backend/kernels_base.f90'],
    **ext_kwargs)


if __name__ == "__main__":
    # explicitly allow installation in user site in development mode
    site.ENABLE_USER_SITE = "--user" in sys.argv[1:]

    use_fortran = os.environ.get("USE_FORTRAN", None)
    if use_fortran:
        setup(ext_modules=[ext_gpfunc, ext_kernels])
    else:
        setup()



from numpy import *
from profit.sur.backend.gp_functions_old import build_K
from profit.sur.backend.gpfunc import gpfunc
from time import time
#import tracemalloc
#tracemalloc.start()

def nu_L2(xa, xb):
    return sum((xa - xb)**2, axis=1)

def kern_sqexp(nu):
    return exp(nu)

def build_numpy(xa, xb, kern, nu, K):
    for ka in arange(xa.shape[0]):
        K[ka, :] = kern(nu(xa[ka, :], xb))

nd = 4
na = 4096
nb = na

l = ones(nd)
xa = random.rand(na, nd)
xb = xa
K = empty((na, nb))

tic = time()
#snapshot1 = tracemalloc.take_snapshot()
build_numpy(xa, xb, kern_sqexp, nu_L2, K)
toc = time() - tic
#snapshot2 = tracemalloc.take_snapshot()
#top_stats = snapshot2.compare_to(snapshot1, 'lineno')
print('Python: ', toc)

K = empty((na, nb), order='F')
tic = time()
build_K(xa, xb, l, K)
toc = time() - tic
print('Fortran: ', toc)

K = empty((na, nb), order='F')
tic = time()
build_numpy(xa, xb, gpfunc.kern_sqexp, nu_L2, K)
toc = time() - tic
print('Call Fortran kernel from Python: ', toc)



import numpy as np
import yaml

with open('config.yaml') as f:
    data = yaml.load(f, Loader=yaml.FullLoader)

Nm = data['Nm'] #mapping time
N = data['N']

sig = data['sig'] # sigma for GP
sig_n = data['sig_n'] # noise for GP

U0 = data['U0']

nm = data['nm'] # how often the map should be applied
dtsymp = data['dtsymp']

qmin = data['qmin']
qmax = data['qmax']
pmin = data['pmin']
pmax = data['pmax']

qminmap = np.array(np.pi - 2.8)
qmaxmap = np.array(np.pi + 1.5)
pminmap = np.array(-2.3)
pmaxmap = np.array(1.8)
Ntest = 3


#%% See also https://i-systems.github.io/teaching/ML/iNotes/15_Autoencoder.html
#%%
import numpy as np
import plotly.express as px
import plotly.graph_objects as go
import pandas as pd
import sklearn as skl
from sklearn.neural_network import MLPRegressor
#%%

nskip = 1000
skip = np.arange(1e6)
skip = np.delete(skip, np.arange(0, 1e6, nskip))

data = pd.read_table(
    '/mnt/c/Users/chral/Dropbox/ipp/paper_algae/mc_out.dat', sep='\s+', skiprows=skip)
names = {
    'mu_0': 'k_alg_growth_max',
    'f_si': 'frac_si_alg_1',
    'lambda_S': 'k_att_shade',
    'K_light': 'k_light_sm',
    'sigma_0': 'k_alg_loss',
    'a': 'coeff_d_loss_2'
}
# %%
indata = data[['k_alg_loss', 'frac_si_alg_1']].values
indata = (indata - np.min(indata,0))/np.max(indata - np.min(indata,0),0)
px.scatter(x=indata[:,0], y=indata[:,1],
    labels={'x': 'x[0]', 'y': 'x[1]'})
# %%
reg = MLPRegressor(hidden_layer_sizes = (4, 8, 4, 1, 4, 8, 4), 
                   activation = 'tanh', 
                   solver = 'lbfgs', 
                   max_iter = 512, 
                   tol = 0.0000001, 
                   verbose = True)

train, test = skl.model_selection.train_test_split(indata)
#%%

reg.fit(train, train)
#%% Plot result on training data
output_eval_train = reg.predict(train)
fig = px.scatter(x=train[:,0], y=train[:,1],
    labels={'x': 'x[0]', 'y': 'x[1]'})
fig.add_scatter(x=output_eval_train[:,0], y=output_eval_train[:,1], mode='markers')
#%% Plot result on test data
output_eval_test = reg.predict(test)
fig = px.scatter(x=test[:,0], y=test[:,1],
    labels={'x': 'x[0]', 'y': 'x[1]'})
fig.add_scatter(x=output_eval_test[:,0], y=output_eval_test[:,1], mode='markers')
# %% Cut the network in half at the hidden 1D layer
#    to get values of the hidden parameter
reg.n_layers_ = ((reg.n_layers_ - 2)+1) // 2 + 1
#%%
ae_parm = reg.predict(train)
fig = go.Figure()
fig.add_scatter(x=ae_parm, y=train[:,0], 
    mode='markers', name='x[0]')
fig.add_scatter(x=ae_parm, y=output_eval_train[:,0], 
    mode='markers', name='x[0] reduced')
fig.add_scatter(x=ae_parm, y=train[:,1], 
    mode='markers', name='x[1]')
fig.add_scatter(x=ae_parm, y=output_eval_train[:,1], 
    mode='markers', name='x[1] reduced')
fig.update_layout(
    xaxis_title = 't (hidden curve parameter)',
    yaxis_title = 'x[0], x[1]')
# %%
ae_parm = reg.predict(test)
fig = go.Figure()
fig.add_scatter(x=ae_parm, y=test[:,0], 
    mode='markers', name='x[0]')
fig.add_scatter(x=ae_parm, y=output_eval_test[:,0], 
    mode='markers', name='x[0] reduced')
fig.add_scatter(x=ae_parm, y=test[:,1], 
    mode='markers', name='x[1]')
fig.add_scatter(x=ae_parm, y=output_eval_test[:,1], 
    mode='markers', name='x[1] reduced')
fig.update_layout(
    xaxis_title = 't (hidden curve parameter)',
    yaxis_title = 'x[0], x[1]')

# %%
outdata = data[['cost_function']].values
#%%



"""
Created: Tue Mar 26 09:45:46 2019
@author: Christopher Albert <albert@alumni.tugraz.at>
"""
import numpy as np
import os
from cffi import FFI

docompile = False

ffi = FFI()

if docompile:
    cwd = os.getcwd()
    
    ffi.set_source('_libuq', '', libraries=['uq'])
    ffi.cdef("""
       void __mod_unqu_MOD_init_uq();
       void __mod_unqu_MOD_pre_uq(double *axi);
       void __mod_unqu_MOD_run_uq();
       void __mod_unqu_MOD_post_uq();
    """)
    
    ffi.compile(verbose=True)

#%%
import _libuq
libuq = _libuq.lib

npol = 3
npar = 2
nall = (npol+2)**npar

axi = np.zeros(nall*npar, dtype=np.float32)

libuq.__mod_unqu_MOD_init_uq()
#libuq.__mod_unqu_MOD_pre_uq(ffi.cast('double*', axi.ctypes.data))
_libuq.lib.__mod_unqu_MOD_run_uq()
#%%

import numpy as np
import matplotlib.pyplot as plt

nt = 1

a0 = np.zeros(nt)
for kt in np.arange(nt):
    data = np.loadtxt('akoef{:04d}.dat'.format(kt+1))
    a0[kt] = data[0,2]
    
plt.plot(a0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb  7 12:26:56 2019

@author: ert
"""

#
# Important: need to do sys.path.append('/path/to/redmod/..') to be able to use this script
#

import numpy as np
import matplotlib.pyplot as plt

from numpy import sqrt, log
from chaospy import (Normal, J, generate_quadrature, orth_ttr, fit_quadrature, 
                     E, Std, descriptives)

#%%
indata = np.genfromtxt('2019-07_run_1998/input_full.txt', names=True)
outdata = np.genfromtxt(
        '2019-07_run_1998/1998_V7_only_Chl_output_UQ_1998_sparse_proc_000.txt',
        skip_header=1)
outdata = outdata[outdata[:,0].argsort()]

#%%


data = np.genfromtxt('2019-07_run_1998/params.txt', skip_header=1, 
                     dtype=('|U64','|U64',float,float))

labels = data['f0']  # parameter labels
mean = data['f2']    # E0
std = data['f3']     # sqrt(Var0)

s = sqrt(log(std**2/mean**2 + 1))
mu = log(mean) - 0.5*s**2

params = []
for k in range(len(mu)):
    params.append(Normal(mu = mu[k], sigma=s[k]))

dist = J(*params)
#%%

nodes, weights = generate_quadrature(4, dist, rule='G', sparse=True)
expansion = orth_ttr(3, dist)

#%%
approx = fit_quadrature(expansion, nodes, weights, outdata)

#%%

F0 = E(approx, dist)
dF = Std(approx, dist)

#%%
plt.figure(figsize=(6, 3))
plt.plot(F0, 'k')
plt.fill_between(range(len(F0)),F0-1.96*dF, F0+1.96*dF, alpha=0.2)  # 95% CI
plt.fill_between(range(len(F0)),F0-0.67*dF, F0+0.67*dF, alpha=0.5)  # 50% CI
plt.grid(True)
plt.xlim(1,80)
plt.ylim(0,300)

#%%
sobol1 = descriptives.sensitivity.Sens_m(approx, dist)

# yielded array([0.49578562, 0.00260017, 0.00284624, 0.00343361, 0.31835111,
#                0.12970096, 0.00513917])

#%%
plt.figure(figsize=(10,5))
plt.plot(sobol1.T[1:,:])
plt.grid(True)
plt.legend(labels)


"""
Created: Tue Jul 30 07:56:52 2019
@author: Christopher Albert <albert@alumni.tugraz.at>
"""
import os
import site

from julia.api import LibJulia, Julia

api = LibJulia.load(julia=os.path.join(site.USER_BASE, 'bin', 'julia-py'))
api.sysimage = 'julia_sysimage.so'
api.init_julia()

#%%

ret = api.jl_eval_string(b"sin(pi)")
print(float(api.jl_unbox_float64(ret)))
